"""Utilities for working with the file system."""
import os
import time
import pathlib
from typing import Optional
import urllib.error
import urllib.request
import zipfile

from absl import logging

import inductiva
from inductiva import types


def find_path_to_package(package_dir: str):
    """Find path to package directory in inductiva library in local computer.

    Args:
        package_dir: Name of package directory.
    """
    return os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", package_dir))


def get_timestamped_path(path: types.Path, sep: str = "-") -> pathlib.Path:
    """Return a path that does not exist by appending a timestamp.

    Args:
        path: Path to a file or directory.

    Returns:
        A path that does not exist by appending the timestamp.
    """
    path = pathlib.Path(path)
    timestamp = time.strftime("%Y-%m-%dT%Hh%Mm%Ss")

    name = f"{path.stem}{sep}{timestamp}"

    return path.with_name(name + path.suffix)


def resolve_path(path: Optional[types.Path]) -> pathlib.Path:
    """Resolve a path relative to the Inductiva package working directory.

    This takes in a path and returns it resolved relative to the
    "inductiva.working_dir" setting.
    Note:
    If the working directory is not set, the current working directory of
    the script is used.
    If the path is absolute, it will override the working directory and be
    returned as is.
    If the path is None, the working directory will be returned.
    Check the tests in "test_files_utils.py" for examples.


    Args:
        path: Path to a file or directory.
    """
    resolved_path = pathlib.Path.cwd()

    if inductiva.working_dir:
        resolved_path = pathlib.Path(inductiva.working_dir)

    if path is not None:
        resolved_path = pathlib.Path(resolved_path, path)

    return resolved_path


def get_sorted_files(data_dir: str,
                     file_format: str = "name",
                     split_token: str = "_"):
    """Returns list of files sorted according to [file_key].

    Order a set of .format files of the form
    ['name_1.format', 'name_2.format',...,'name_10.format',
    ...,'name_n.format'].

    The default sorting methods for list, list.sort()
    or sorted(list), order 'name_10.format' before 'name_2.format',
    which is not representative of the time series.

    In this function we sort according to the number..
    """

    if not os.path.exists(data_dir):
        raise IOError(f"Directory '{data_dir}' does not exist.")

    # Get a list of the files in the data directory.
    files = os.scandir(data_dir)

    # The files have file_format extension.
    files = [
        file for file in files if pathlib.Path(file.path).suffix == file_format
    ]

    # Sort the files to be read according to [file_key].
    def get_alphanum_key(file):
        file_name = pathlib.Path(file.path).stem
        file_name_splits = file_name.split(split_token)
        file_key = file_name_splits[-1]
        return int(file_key)

    files = sorted(files, key=get_alphanum_key)

    return files


def _unzip(zip_path: pathlib.Path, dest_path: pathlib.Path):
    """Unzip a zip archive.

    If the ZIP contains one file, dest_path will be the path of that single
    file. If the ZIP contains multiple files, dest_path will be the path
    of the directory containing the files.
    """
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        files = zip_ref.namelist()
        if len(files) == 1:
            # Unzip the file to the directory of dest_path
            unzipped_file_path = pathlib.Path(
                zip_ref.extract(files[0], dest_path.parent))
            pathlib.Path(unzipped_file_path).rename(dest_path)
        else:
            dest_path.mkdir(parents=True, exist_ok=True)
            zip_ref.extractall(dest_path.parent)

    # Rename the file to dest_path
    zip_path.unlink()


def download_from_url(url: str, save_dir: Optional[str] = None) -> str:
    """Download a file from an URL.

    If the file is a ZIP archive containing a single file, the method
    will extract the file and rename it to `local_file_path`. If the ZIP
    contains more than one file, a ValueError will be raised.

    Args:
        url: The URL to download the file from.
        local_file_path: Name attributed to the file to download.
            If None is passed, it attributes the name given to
            the file on the url.
        save_dir: The directory to save the file to. If None
            is passed, this will download to the current working
            directory. If the save_dir passed does not exist, this
            method will try to create it.
    Returns:
        The path to the downloaded file.
    """
    local_path = url.split("/")[-1]
    local_path = pathlib.Path(local_path)

    if save_dir is not None:
        local_path = pathlib.Path(save_dir, local_path)

    local_path = resolve_path(local_path)

    if not local_path.parent.exists():
        local_path.parent.mkdir(parents=True)

    try:
        downloaded_to, headers = urllib.request.urlretrieve(url)
    except urllib.error.URLError as url_error:
        logging.error("Could not download file from %s", url)
        raise url_error

    # File was downloaded to a temporary file
    downloaded_to = pathlib.Path(downloaded_to)
    is_zip = headers.get_content_type() == "application/zip"

    if is_zip:
        # Unzip the ZIP archive containing a single zip file to the
        # correct path

        if local_path.suffix == ".zip":  # Remove the .zip extension
            local_path = local_path.with_suffix("")

        _unzip(downloaded_to, local_path)
    else:
        # Rename the file to the correct path
        downloaded_to.rename(local_path)

    return str(local_path.absolute())



"""Connect tasks to simulation output classes via method_name."""
from inductiva import fluids, molecules, coastal

OUTPUT_CONSTS = {
    "wind_tunnel": {
        "output_class":
            fluids.WindTunnelOutput,
        "default_files": [
            "pressure_field.vtk", "streamlines.vtk", "stdout.txt", "stderr.txt",
            "force_coefficients.csv", "xy_flow_slice.vtk", "xz_flow_slice.vtk",
            "yz_flow_slice.vtk", "constant/triSurface/object.obj"
        ]
    },
    "wind_terrain": {
        "output_class": fluids.SteadyStateOutput,
        "default_files": None
    },
    "heat_sink": {
        "output_class": fluids.HeatSinkOutput,
        "default_files": None
    },
    "dam_break": {
        "output_class": fluids.SPHSimulationOutput,
        "default_files": None
    },
    "fluid_block": {
        "output_class": fluids.SPHSimulationOutput,
        "default_files": None
    },
    "fluid_tank": {
        "output_class": fluids.FluidTankOutput,
        "default_files": None
    },
    "coastal_area": {
        "output_class": coastal.CoastalAreaOutput,
        "default_files": None
    },
    "protein_solvation": {
        "output_class":
            molecules.ProteinSolvationOutput,
        "default_files": [
            "protein.gro", "compressed_trajectory.xtc", "solvated_protein.tpr",
            "topol.top", "rmsf_values.npy"
        ]
    },
    "mdwater_box": {
        "output_class": molecules.MDWaterBoxOutput,
        "default_files": None
    },
    "stellarators": {
        "output_class": None,
        "default_files": None
    },
    "deformable_plate": {
        "output_class": None,
        "default_files": None
    }
}



"""Post-processing and visualization utilities of the heat sink scenario."""

import os
import pathlib
from typing import Literal, Sequence

try:
    import pyvista as pv
except ImportError:
    pv = None

from inductiva.types import Path
from inductiva.utils import files, optional_deps


class HeatSinkOutput:
    """Heat sink simulation output."""

    def __init__(self, sim_output_path: Path):
        """Initializes a `HeatSinkOutput` object.

        Args:
            sim_output_path: Path to simulation output files.
        """

        self.sim_output_path = sim_output_path

    @optional_deps.needs_fluids_extra_deps
    def render(
        self,
        movie_path: Path = "movie.mp4",
        fps: int = 10,
        cmap: str = "viridis",
        clim: Sequence[float] = (280, 300),
        view: Literal["isometric", "front", "rear", "top",
                      "side"] = "isometric",
    ):
        """Renders temporal evolution of the temperature.
        
        A movie is produced representing the temporal evolution of the
        temperature in the heat sink and in the surrounding air flow.
         
        The movie is saved in the `movie_path` location.

        Args:
            movie_path: Path to the movie file.
            fps: Number of frames per second to use in the movie.
            cmap: Colormap used to represent temperature.
            clim: Colorbar limits.
            view: View to use in the movie.
        """

        movie_path = files.resolve_path(movie_path)

        # The OpenFOAM data reader from PyVista requires that a file named
        # "foam.foam" exists in the simulation output directory.
        # Create this file if it does not exist.
        foam_file_path = os.path.join(self.sim_output_path, "foam.foam")
        pathlib.Path(foam_file_path).touch(exist_ok=True)

        reader = pv.OpenFOAMReader(foam_file_path)

        plotter = pv.Plotter(off_screen=True)

        # Set camera position for a nice view.
        if view == "isometric":
            plotter.view_vector([-0.67, 0.49, -0.58], viewup=[0.31, 0.90, 0.29])
            zoom_factor = 1.4
            slice_normal = (-1, 0, 0)
        elif view == "front":
            plotter.view_xy(negative=True)
            zoom_factor = 1.6
            slice_normal = (0, 0, -1)
        elif view == "rear":
            plotter.view_xy()
            zoom_factor = 1.6
            slice_normal = (0, 0, 1)
        elif view == "top":
            plotter.view_zx()
            zoom_factor = 1.2
            slice_normal = (0, -1, 0)
        elif view == "side":
            plotter.view_zy()
            zoom_factor = 1.2
            slice_normal = (-1, 0, 0)
        else:
            raise ValueError("Invalid view.")

        plotter.camera.zoom(zoom_factor)

        plotter.open_movie(movie_path, framerate=fps)

        for idx in range(1, reader.number_time_points):
            reader.set_active_time_point(idx)

            mesh = reader.read()

            fins_data = mesh["fins"]["boundary"]
            fluid_data = mesh["fluid"]["internalMesh"]

            fluid_data_slice = fluid_data.slice(normal=slice_normal,
                                                origin=(0, 0.02, 0))

            plotter.add_mesh(fins_data,
                             scalars="T",
                             cmap=cmap,
                             clim=clim,
                             show_scalar_bar=False)

            plotter.add_mesh(fluid_data_slice,
                             scalars="T",
                             cmap=cmap,
                             clim=clim,
                             show_scalar_bar=False)

            # Create custom color bar for temperature.
            plotter.add_scalar_bar(
                "Temperature [K]",
                position_x=0.45,
                width=0.5,
                title_font_size=18,
                label_font_size=18,
                fmt="%.0f",
                font_family="arial",
            )

            # Add time label.
            plotter.add_title(f"Time = {reader.active_time_value} [s]",
                              font_size=10)

            plotter.show_axes()

            plotter.write_frame()

        plotter.close()



"""Post process SPH simulation outputs."""
import os

from inductiva.utils import visualization, optional_deps
from inductiva.types import Path


class SPHSimulationOutput:
    """Post process SPH simulation outputs."""

    def __init__(self, sim_output_path: Path):
        """Initializes a `SimulationOutput` object.

        Args:
            sim_output_path: Path to simulation output files.
            """
        self.sim_output_dir = sim_output_path

    @optional_deps.needs_fluids_extra_deps
    def render(self,
               virtual_display: bool = False,
               movie_path: str = "movie.mp4",
               fps: int = 10,
               color: str = "blue"):
        """Render the simulation as a movie.

        Args:
        fps: Number of frames per second to use in the movie. Renders a
            subset of the vtk files to create the movie.
            Default: 10.
        color: The color of the markers in the simulation.
          If None, the default color is used.
        """

        vtk_dir = os.path.join(self.sim_output_dir, "vtk")

        visualization.create_movie_from_vtk(vtk_dir,
                                            movie_path,
                                            virtual_display=virtual_display,
                                            camera=[(3., 3., 2.), (0., 0., 0.),
                                                    (1., 1., 2.)],
                                            fps=fps,
                                            color=color)



"""Base class for scenarios."""

from abc import ABC, abstractmethod
import io
import tempfile
from typing import Optional, Union

from inductiva import resources
from inductiva.types import Path
from inductiva.simulators import Simulator
import json


class Scenario(ABC):
    """Base class for scenarios."""
    valid_simulators = []

    @abstractmethod
    def create_input_files(self, simulator: Simulator, input_dir: Path):
        """To be implemented in subclasses."""
        pass

    def read_commands_from_file(self, commands_file: Union[str, io.StringIO]):
        "Read list of commands from commands.json file"
        if isinstance(commands_file, str):
            with open(commands_file, "r", encoding="utf-8") as f:
                return json.load(f)

        # Make sure already opened file is read from the beginning
        commands_file.seek(0)
        return json.load(commands_file)

    def validate_simulator(self, simulator: Simulator):
        """Checks if the scenario can be simulated with the given simulator."""
        if type(simulator) not in self.valid_simulators:
            raise ValueError(
                f"Simulator not supported for `{self.__class__.__name__}` "
                "scenario.")

    def simulate(
        self,
        simulator: Simulator,
        machine_group: Optional[resources.MachineGroup] = None,
        run_async: bool = False,
        **kwargs,
    ):
        """Simulates the scenario synchronously."""
        self.validate_simulator(simulator)

        with tempfile.TemporaryDirectory() as input_dir:
            self.create_input_files(simulator, input_dir)

            return simulator.run(
                input_dir,
                machine_group=machine_group,
                run_async=run_async,
                **kwargs,
            )



"""Bathymetry class and utilities."""

import random
from typing import Optional, Sequence, Tuple, Union

from absl import logging

import numpy as np
try:
    import matplotlib
    import utm
except ImportError:
    matplotlib = None
    utm = None

import inductiva
from inductiva.utils import optional_deps


class Bathymetry:
    """Represents a bathymetric profile.

    A bathymetric profile defines the depth of the sea bottom as a function of
    space, here described in Cartesian coordinates (x, y).

    Here, a bathymetry is represented with:
    - a 1D array of depths, in meters, measured at arbitrary points in space.
      Positive depths are below the water level.
    - two 1D arrays representing the x and y coordinates of the points where the
      depths are defined, in meters.
    """

    @optional_deps.needs_coastal_extra_deps
    def __init__(
        self,
        depths: np.ndarray,
        x: np.ndarray,
        y: np.ndarray,
    ):
        """Initializes a `Bathymetry` object.

        Args:
            depths: A 1D array with the depths, in meters.
            x: A 1D array with the x coordinates of the points where depths are
              defined, in meters.
            y: Same as `x`, but for the y coordinates.
        """
        self.depths = depths
        self.x = x
        self.y = y

    @classmethod
    @optional_deps.needs_coastal_extra_deps
    def from_bot_file(
        cls,
        bot_file_path: str,
        x_range: Sequence[float],
        y_range: Sequence[float],
    ):
        """Creates a `Bathymetry` object from a bot file.

        The depth values are read from a bot file, i.e. a text file with a 2D
        array with the depths, in meters. The first and second dimensions of the
        array in the text file (i.e. rows and columns) correspond to the x and y
        directions, respectively.

        Args:
            text_file_path: Path to the text file.
            x_range: The range of x values, in meters.
            y_range: The range of y values, in meters.
        """

        depths = np.loadtxt(bot_file_path)

        x, y = np.meshgrid(np.linspace(*x_range, depths.shape[0]),
                           np.linspace(*y_range, depths.shape[1]),
                           indexing="ij")

        return cls(depths.flatten(), x.flatten(), y.flatten())

    @classmethod
    @optional_deps.needs_coastal_extra_deps
    def from_ascii_xyz_file(
        cls,
        ascii_xyz_file_path: str,
        remove_offset: bool = True,
    ):
        """Creates a `Bathymetry` object from an ASCII XYZ file.

        ASCII XYZ files store bathymetric data in a table where each line
        corresponds to a location (latitude, longitude pair). Several columns
        are available to characterize the depth at each location, namely:
        - longitude, in decimal degrees, e.g. 52.07334567;
        - latitude, in decimal degrees, e.g. 3.06033283;
        - the minum depth, in meters, e.g. 35.81;
        - the maximum depth, in meters, e.g. 35.81;
        - average depth (over a set of measurements), in meters, e.g. 35.81.
        - the standard deviation of the depth (over a set of measurements), in
          meters, e.g. 35.81;

        For more information, see
        https://emodnet.ec.europa.eu/sites/emodnet.ec.europa.eu/files/public/20171127_DTM_exchange_format_specification_v1.6.pdf.

        In this method, we:
        - load the latitude and longitude values from the file, and convert
          them to (x, y) UTM coordinates (see
          https://en.wikipedia.org/wiki/Universal_Transverse_Mercator_coordinate_system
          for more information).
        - use average depth as the local depth value.

        Args:
            ascii_xyz_file_path: Path to the ASCII XYZ file.
            remove_offset: Whether to remove the offset of the x and y
              coordinates. If `True`, the minimum x and y values are subtracted
              from all x and y values, respectively.
        """

        # Load only the first, second and fifth columns of the file,
        # corresponding to longitude, latitude and average depth, respectively.
        bathymetry_data = np.loadtxt(ascii_xyz_file_path,
                                     usecols=(0, 1, 4),
                                     delimiter=";")

        x_lon = bathymetry_data[:, 0]
        y_lat = bathymetry_data[:, 1]
        depths = bathymetry_data[:, 2]

        x, y, _, _ = utm.from_latlon(y_lat, x_lon)

        if remove_offset:
            x = x - x.min()
            y = y - y.min()

        return cls(depths, x, y)

    @classmethod
    @optional_deps.needs_coastal_extra_deps
    def from_random_depths(
        cls,
        x_range: Sequence[float],
        y_range: Sequence[float],
        x_num: int,
        y_num: int,
        max_depth: float = 10,
        initial_roughness: float = 1,
        roughness_factor: float = 0.5,
        percentile_above_water: float = 20,
        random_seed=None,
    ):
        """Creates a `Bathymetry` object with random depths.

        The depths of the corners of the grid are chosen according to a maximum
        depth value `max_depth` and a percentile of the domain that must be
        above water `percentile_above_water`. The corners on the lower x
        boundary (East) are assumed to be below water (i.e. have 0 < depths <
        `max_depth`). The corners on the upper x boundary (West) are assumed to
        be above water (i.e. have - `max_depth` < depths < 0).

        Args:
            x_range: The range of x values, in meters.
            y_range: The range of y values, in meters.
            x_num: Number of grid points in the x direction.
            y_num: Number of grid points in the y direction.
            max_depth: Maximum depth value, in meters.
            initial_roughness: Initial roughness value, in meters. Controls the
              initial range of randomness of the Diamond-Square algorithm.
            roughness_factor: Roughness factor. Must be between 0 and 1.
              Controls the rate at which the range of randomness of the
              Diamond-Square algorithm decreases over iterations.
            percentile_above_water: Percentile of the depths that must be above
              water. Must be between 0 and 100.
            random_seed: Random seed to use.
        """

        random.seed(random_seed)

        corner_values = [
            random.uniform(0, max_depth),
            random.uniform(0, max_depth),
            random.uniform(-max_depth, 0),
            random.uniform(-max_depth, 0),
        ]

        depths = inductiva.generative.procedural.generate_random_map_level(
            x_num,
            y_num,
            corner_values,
            initial_roughness,
            roughness_factor,
            random_seed=random_seed,
        )

        depths = inductiva.generative.procedural.adjust_map_level(
            depths, percentile_above_water)

        x, y = np.meshgrid(np.linspace(*x_range, x_num),
                           np.linspace(*y_range, y_num),
                           indexing="ij")

        return cls(depths.flatten(), x.flatten(), y.flatten())

    @optional_deps.needs_coastal_extra_deps
    def to_bot_file(self, bot_file_path: str, depths_grid: np.ndarray):
        """Writes the bathymetry to a bot file.

        The depth values are interpolated to a regular grid and written to a bot
        file. The grid size is determined automatically from the range and the
        number of unique x and y values.

        Args:
            text_file_path: Path to the text file.
        """

        np.savetxt(bot_file_path, depths_grid)

    @property
    @optional_deps.needs_coastal_extra_deps
    def x_range(self) -> Tuple[float]:
        """Returns the range of x values."""

        return (np.min(self.x), np.max(self.x))

    @property
    @optional_deps.needs_coastal_extra_deps
    def y_range(self) -> Tuple[float]:
        """Returns the range of y values."""

        return (np.min(self.y), np.max(self.y))

    @optional_deps.needs_coastal_extra_deps
    def x_ptp(self) -> float:
        """Returns the peak-to-peak range (max - min) of x values."""

        return np.ptp(self.x)

    @optional_deps.needs_coastal_extra_deps
    def y_ptp(self) -> float:
        """Returns the peak-to-peak range (max - min) of y values."""

        return np.ptp(self.y)

    @optional_deps.needs_coastal_extra_deps
    def x_num(self) -> int:
        """Returns the length of unique x values."""

        return int(len(self.x_uniques()))

    @optional_deps.needs_coastal_extra_deps
    def y_num(self) -> int:
        """Returns the length of unique y values."""

        return int(len(self.y_uniques()))

    @optional_deps.needs_coastal_extra_deps
    def x_uniques(self, sort: bool = False) -> np.ndarray:
        """Returns the unique x values.

        Args:
            sort: Whether to sort the unique values.
        """
        x_uniques = np.unique(self.x)
        if sort:
            x_uniques = np.sort(x_uniques)
        return x_uniques

    @optional_deps.needs_coastal_extra_deps
    def y_uniques(self, sort: bool = False) -> np.ndarray:
        """Returns the unique y values.

        Args:
            sort: Whether to sort the unique values.
        """
        y_uniques = np.unique(self.y)
        if sort:
            y_uniques = np.sort(y_uniques)
        return y_uniques

    @optional_deps.needs_coastal_extra_deps
    def depths_grid(self) -> np.ndarray:
        """Returns the depths as a 2D grid."""

        return self.depths.reshape((self.x_num(), self.y_num()))

    @optional_deps.needs_coastal_extra_deps
    def crop(self,
             x_range: Sequence[float],
             y_range: Sequence[float],
             remove_offset=True):
        """Crops the bathymetry to a given range of x and y values.

        Args:
            x_range: The range of x values, in meters.
            y_range: The range of y values, in meters.
        """
        x_min, x_max = x_range
        y_min, y_max = y_range

        mask = (self.x >= x_min) & (self.x <= x_max) & \
               (self.y >= y_min) & (self.y <= y_max)

        if np.sum(mask) == 0:
            raise ValueError(
                "The bathymetry cannot be cropped because no points are "
                "defined in the given coordinate ranges.")

        x = self.x[mask]
        y = self.y[mask]
        depths = self.depths[mask]

        if remove_offset:
            x = x - x.min()
            y = y - y.min()

        return Bathymetry(
            depths=depths,
            x=x,
            y=y,
        )

    @optional_deps.needs_coastal_extra_deps
    def is_uniform_grid(self) -> bool:
        """Determines whether the bathymetry is defined on a uniform grid."""

        x_uniques_diffs = np.diff(self.x_uniques(sort=True))
        y_uniques_diffs = np.diff(self.y_uniques(sort=True))

        return (np.unique(x_uniques_diffs.round(decimals=2)).size == 1 and
                np.unique(y_uniques_diffs.round(decimals=2)).size == 1)

    @optional_deps.needs_coastal_extra_deps
    def plot(
        self,
        show: bool = False,
        cmap: str = "coolwarm",
        clim: Optional[Tuple[float]] = None,
        path: Optional[str] = None,
        x_resolution: float = 10,
        y_resolution: float = 10,
        resize: bool = False,
        threshold_distance: float = 20,
    ) -> Union["matplotlib.axes.Axes", None]:
        """Plots the bathymetry.

        The bathymetry is represented as a 2D map of depths, with the x and y
        coordinates of the points where the depths are defined in the axes. The
        plot shows the height of the various points in the bathymetry is
        defined, with the opposite signal convention of the bathymetry (i.e,
        the bellow water level points are given by negative height instead of
        positive depth).

        The bathymetry data is plotted with a color plot on a grid with uniform
        spacing in the x and y directions. If the user does not specify the
        resolution, the plotted mesh has the bathymetry resolution. If
        necessary, the user can control the spacing in the x and y directions
        using the `x_resolution` and `y_resolution` arguments.

        For non-uniform grids or if the resoltuion is specified, the data is
        interpolated from the points where the bathymetry is defined
        to the resized uniform grid using linear interpolation.

        Points on the uniform grid at a distance larger than a threshold
        distance `threshold_distance` from points where the bathymetry is
        defined are omitted.

        The plot is produced with matplotlib.

        Args:
            cmap: Colormap to use. Defaults to the matplotlib default colormap.
            clim: Range of depth values to represent in colors. If `None`, the
              range is determined automatically from the minumum and maximum
              depth. Depth values below or above this range are represented with
              the minimum or maximum colors, respectively.
            path: Path to save the plot. If `None`, the plot is not saved, and
              the matplotlib `Axes` object is returned instead.
            x_resolution: Resolution, in meters, of the plotting grid in the x
              direction. Only used if the bathymetry is not uniform or if we
              are resizing it.
            y_resolution: Resolution, in meters, of the plotting grid in the y
              direction. Only used if the bathymetry is not uniform or if we
              are resizing it.
            resize: Whether to resize the bathymetry before plotting it. When
                `False`, the bathymetry is plotted on the original grid if
                it is already uniform.
            threshold_distance: Threshold distance to filter out points on the
              uniform grid that are far from points where the bathymetry is
              defined.
        """

        if self.is_uniform_grid() and not resize:
            x_size = self.x_num()
            y_size = self.y_num()

            depths_grid = self.depths.reshape((x_size, y_size))

        else:
            logging.info("Plotting the bathymetry on resized uniform grid...")

            depths_grid, _ = self._interpolate_to_uniform_grid(
                x_resolution,
                y_resolution,
                threshold_distance=threshold_distance,
                nullable=True,
            )

        # Plot the bathymetry.
        extent = (
            self.x_range[0],
            self.x_range[1],
            self.y_range[0],
            self.y_range[1],
        )

        fig = matplotlib.pyplot.figure()
        ax = fig.add_subplot()

        im = ax.imshow(
            -1 * depths_grid.transpose(),
            cmap=cmap,
            clim=clim,
            origin="lower",
            extent=extent,
        )

        ax.set(
            aspect="equal",
            xlim=self.x_range,
            ylim=self.y_range,
            xlabel="$x$ [m]",
            ylabel="$y$ [m]",
        )

        fig.colorbar(im, ax=ax, label="Height [m]")

        if path is not None:
            fig.savefig(path)
            matplotlib.pyplot.close(fig)

        else:
            if show:
                matplotlib.pyplot.show()
            return ax

    @optional_deps.needs_coastal_extra_deps
    def to_uniform_grid(
        self,
        x_resolution: float = 2,
        y_resolution: float = 2,
        fill_value: Optional[Union[float, str]] = None,
    ):
        """Converts the bathymetry to a uniform grid.

        The bathymetry is interpolated to a grid with uniform resolution (i.e.
        spacing) in the x and y directions. Linear interpolation is used to
        interpolate the bathymetry to the grid.

        Grid points for which no interpolation is possible may be filled with
        a constant or with the nearest depth value.

        Args:
            x_resolution: Resolution, in meters, of the grid in the x direction.
            y_resolution: Resolution, in meters, of the grid in the y direction.
            fill_value: Value to fill the grid points for which no interpolation
              is possible. If "nearest", the nearest depth value is used.
        """

        depths_grid, (x_grid, y_grid) = self._interpolate_to_uniform_grid(
            x_resolution, y_resolution, fill_value=fill_value, nullable=False)

        return Bathymetry(depths=depths_grid.flatten(),
                          x=x_grid.flatten(),
                          y=y_grid.flatten())

    @optional_deps.needs_coastal_extra_deps
    def _interpolate_to_uniform_grid(
        self,
        x_resolution: float,
        y_resolution: float,
        threshold_distance: Optional[float] = None,
        fill_value: Optional[Union[float, str]] = None,
        nullable: bool = False,
    ):
        """Interpolates the bathymetry to a uniform grid.

        Args:
            x_resolution: Resolution, in meters, of the grid in the x direction.
            y_resolution: Resolution, in meters, of the grid in the y direction.
            threshold_distance: Threshold distance to filter out points on the
              uniform grid that are far from points where the bathymetry is
              defined.
            fill_value: Value to fill the grid points for which no interpolation
              is possible. If "nearest", the nearest depth value is used.
            nullable: Whether to allow the bathymetry to be undefined in some
              grid points. If `False`, an error is raised if the bathymetry is
              undefined in one or more grid points.

        Returns:
            depths_grid: A 2D array with the depths on the uniform grid.
            (x_grid, y_grid): The x and y coordinates of the points where the
              depths are defined, respectively.
        """
        # Determine grid size based on ranges and resolution.
        x_size = int(self.x_ptp() / x_resolution)
        y_size = int(self.y_ptp() / y_resolution)

        logging.info(
            "Interpolating the bathymetry to a grid of dimensions:\n"
            "- grid resolution %.2f x %.2f (m x m) \n"
            "- grid size %d x %d", x_resolution, y_resolution, x_size, y_size)

        # Create uniform grid for interpolation.
        (x_grid, y_grid) = inductiva.utils.grids.get_meshgrid(
            x_range=self.x_range,
            y_range=self.y_range,
            x_num=x_size,
            y_num=y_size,
        )

        depths_grid = inductiva.utils.interpolation.interpolate_to_coordinates(
            coordinates=(self.x, self.y),
            values=self.depths,
            interpolation_coordinates=(x_grid, y_grid),
            method="linear",
            threshold_distance=threshold_distance)

        nan_mask = np.isnan(depths_grid)

        if fill_value is not None:
            if fill_value == "nearest":
                nearest_depths_grid = \
                    inductiva.utils.interpolation.interpolate_to_coordinates(
                        coordinates=(self.x, self.y),
                        values=self.depths,
                        interpolation_coordinates=(x_grid, y_grid),
                        method="nearest")

                depths_grid[nan_mask] = nearest_depths_grid[nan_mask]
            else:
                depths_grid[nan_mask] = fill_value

        nan_mask = np.isnan(depths_grid)

        if not nullable and np.sum(nan_mask) > 0:
            raise ValueError(
                "The bathymetry cannot be interpolated to a uniform grid "
                "because depths are not defined in one or more grid points.")

        return depths_grid, (x_grid, y_grid)



#pylint: disable=missing-module-docstring
from inductiva.tasks.task import Task
from inductiva.tasks.run_simulation import run_simulation
from inductiva.tasks.methods import get, list  # pylint: disable=redefined-builtin



"""Wind tunnel scenario to run an object over an air flow."""

from dataclasses import dataclass
import enum
from functools import singledispatchmethod
import os
import shutil
import io
from typing import Optional, List, Literal

from absl import logging
import numpy as np

from inductiva import tasks, resources, types, simulators, scenarios, utils

SCENARIO_TEMPLATE_DIR = os.path.join(utils.templates.TEMPLATES_PATH,
                                     "wind_tunnel")
OPENFOAM_TEMPLATE_INPUT_DIR = "openfoam"
FILES_SUBDIR = "files"
COMMANDS_TEMPLATE_FILE_NAME = "commands.json.jinja"


@dataclass
class MeshResolution(enum.Enum):
    """Sets particle radius according to resolution."""
    HIGH = [5, 6]
    MEDIUM = [4, 5]
    LOW = [3, 4]
    VERY_LOW = [2, 3]


class WindTunnel(scenarios.Scenario):
    """Physical scenario of a configurable wind tunnel simulation.

    A wind tunnel is a tool used in aerodynamic research to study the
    effects of air moving past solid objects. Here, the tunnel consists
    of a box object in 3D space (x, y, z) space, where air flows in the
    positive x-direction with a certain velocity.

    An arbitrary object is placed within the tunnel, sucht that air flows
    around it, as illustrated in the schematic below:
    |--------------------------------|
    |->          _____               |
    |->        _/     |              |
    |->_______|_o___O_|______________|

    This scenario solves steady-state continuity and momentum equations
    (time-independent) with incompressible flow.
    The simulation solves the time-independent equations for several
    time steps, based on the state of the previous one. The end goal is
    to determine the steady-state of the system, i.e., where the flow
    does not change in time anymore.

    Currently, the following variables are fixed:
    - The fluid being inject is air.
    - The flow is incompressible (this restricts the max air velocity).
    - Air only flows in the positive x-direction.
    - Some post-processing of the data occurs at run-time: streamlines,
    pressure_field, cutting planes and force coefficients.
    """

    valid_simulators = [simulators.OpenFOAM]

    def __init__(self,
                 flow_velocity: List[float] = None,
                 domain: Optional[dict] = None):
        """Initializes the `WindTunnel` conditions.

        Args:
            flow_velocity (List): Velocity of the air flow (m/s).
                The maximum velocity magnitude allowed is 100 m/s
            domain (dict): List containing the lower and upper boundary of
                the wind tunnel in each (x, y, z) direction (m). It is the
                natural description with the default OpenFOAM simulator.
        """
        if flow_velocity is None:
            logging.info("Using a default flow velocity: [30, 0, 0].")
            self.flow_velocity = [30, 0, 0]
        elif len(flow_velocity) != 3:
            raise ValueError("`flow_velocity` must have 3 values.")
        elif np.linalg.norm(np.array(flow_velocity)) > 100:
            raise ValueError("The `flow_velocity` magnitude is too high,"
                             " it must be less than 100 m/s.")
        else:
            self.flow_velocity = flow_velocity

        if domain is None:
            logging.info("Using a default domain: `{\"x\":"
                         "[-5, 15], \"y\": [-4, 4], \"z\": [0, 8]}`.")
            self.domain = {"x": [-5, 15], "y": [-4, 4], "z": [0, 8]}
        elif not isinstance(domain, dict):
            raise ValueError(
                "`domain` must be a dictionary of the type `{\"x\":"
                "[-5, 15], \"y\": [-4, 4], \"z\": [0, 8]}`.")
        else:
            self.domain = domain

    def simulate(self,
                 simulator: simulators.Simulator = simulators.OpenFOAM(),
                 machine_group: Optional[resources.MachineGroup] = None,
                 run_async: bool = False,
                 object_path: Optional[types.Path] = None,
                 num_iterations: float = 100,
                 resolution: Literal["high", "medium", "low"] = "medium",
                 n_cores: int = 2) -> tasks.Task:
        """Simulates the wind tunnel scenario synchronously.

        Args:
            simulator: Simulator used to simulate the scenario.
                Valid simulators: OpenFOAM.
            object_path: Path to object inserted in the wind tunnel.
            output_dir: Path to the directory where the simulation output
                is downloaded when running synchronously.
            num_iterations: Number of iterations of the simulator.
            n_cores: Number of cores to use for the simulation.
            resolution: Level of detail of the mesh used for the simulation.
                Options: "high", "medium" or "low".
            machine_group: The machine group to use for the simulation.
        """
        simulator.override_api_method_prefix("wind_tunnel")

        if object_path:
            self.object_path = utils.files.resolve_path(object_path)
        else:
            logging.info("WindTunnel is empty. Object path not specified.")

        self.num_iterations = num_iterations
        self.n_cores = n_cores
        self.resolution = MeshResolution[resolution.upper()].value

        commands = self.get_commands()

        task = super().simulate(simulator,
                                machine_group=machine_group,
                                run_async=run_async,
                                n_cores=n_cores,
                                commands=commands)

        return task

    def get_commands(self):
        """Returns the commands for the simulation."""

        commands_template_path = os.path.join(SCENARIO_TEMPLATE_DIR,
                                              OPENFOAM_TEMPLATE_INPUT_DIR,
                                              COMMANDS_TEMPLATE_FILE_NAME)

        inmemory_file = io.StringIO()
        utils.templates.replace_params(
            template_path=commands_template_path,
            params={"n_cores": self.n_cores},
            output_file=inmemory_file,
        )
        commands = self.read_commands_from_file(inmemory_file)

        return commands

    @singledispatchmethod
    def create_input_files(self, simulator: simulators.Simulator):
        pass


@WindTunnel.create_input_files.register
def _(self, simulator: simulators.OpenFOAM, input_dir):  # pylint: disable=unused-argument
    """Creates OpenFOAM simulation input files."""

    # The WindTunnel with OpenFOAM requires changing multiple files
    template_files_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                      OPENFOAM_TEMPLATE_INPUT_DIR, FILES_SUBDIR)

    # Copy all files from the template dir to the input directory
    shutil.copytree(template_files_dir,
                    input_dir,
                    dirs_exist_ok=True,
                    symlinks=True)

    utils.templates.batch_replace_params(
        templates_dir=input_dir,
        template_filenames=[
            os.path.join("0", "include",
                         "initialConditions_template.openfoam.jinja"),
            os.path.join("system", "controlDict_template.openfoam.jinja"),
            os.path.join("system", "blockMeshDict_template.openfoam.jinja"),
            os.path.join("system", "decomposeParDict_template.openfoam.jinja"),
            os.path.join("system", "snappyHexMeshDict_template.openfoam.jinja")
        ],
        params={
            "flow_velocity": self.flow_velocity,
            "num_iterations": self.num_iterations,
            "n_cores": self.n_cores,
            "domain": self.domain,
            "resolution": self.resolution,
        },
        output_filename_paths=[
            os.path.join(input_dir, "0", "include", "initialConditions"),
            os.path.join(input_dir, "system", "controlDict"),
            os.path.join(input_dir, "system", "blockMeshDict"),
            os.path.join(input_dir, "system", "decomposeParDict"),
            os.path.join(input_dir, "system", "snappyHexMeshDict")
        ],
        remove_templates=True,
    )

    # Add object path to its respective place
    object_dir = os.path.join(input_dir, "constant", "triSurface")
    os.mkdir(object_dir)
    shutil.copy(self.object_path, os.path.join(object_dir, "object.obj"))



"""OpenFOAM module of the API for fluid dynamics."""
from typing import Optional, List

from inductiva import types, tasks, resources, simulators


class OpenFOAM(simulators.Simulator):
    """Class to invoke a generic OpenFOAM simulation on the API."""

    def __init__(self):
        super().__init__()
        self.api_method_name = "fvm.openfoam.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        commands: List[dict],
        machine_group: Optional[resources.MachineGroup] = None,
        n_cores: int = 2,
        run_async: bool = False,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            commands: List of commands to run using the OpenFOAM simulator.
            n_cores: Number of MPI cores to use for the simulation.
            other arguments: See the documentation of the base class.
        """
        return super().run(input_dir,
                           machine_group=machine_group,
                           n_cores=n_cores,
                           commands=commands,
                           run_async=run_async)



"""Manage running/completed tasks on the Inductiva API."""
import pathlib
import contextlib
import time
import json
from absl import logging
from typing import Dict, Any, List, Optional
from typing_extensions import TypedDict
import datetime
import inductiva
from inductiva.client import models
from inductiva import api
from inductiva.client.apis.tags import tasks_api
from inductiva.utils import files
from inductiva.utils import data
from inductiva.utils import output_contents
from inductiva import types
from inductiva import output_consts

import warnings

_TASK_TERMINAL_STATUSES = {
    models.TaskStatusCode.SUCCESS,
    models.TaskStatusCode.FAILED,
    models.TaskStatusCode.KILLED,
    models.TaskStatusCode.EXECUTERFAILED,
    models.TaskStatusCode.EXECUTERTERMINATED,
    models.TaskStatusCode.SPOTINSTANCEPREEMPTED,
}


class Task:
    """Represents a running/completed task on the Inductiva API.

    Example usage:
        task = scenario.simulate_async(...)
        final_status = task.wait()
        info = task.get_info() # dictionary with info about the task
        task.download_outputs(
            filenames=["file1.txt", "file2.dat"] # download only these files
        )

    Attributes:
        id: The task ID.
        _api: Instance of TasksApi (from the generated API client).
    """

    def __init__(self, task_id: str):
        """Initialize the instance from a task ID."""
        self.id = task_id
        self._api = tasks_api.TasksApi(api.get_client())
        self._info = None
        self._status = None

    @classmethod
    def from_api_info(cls, info: Dict[str, Any]) -> "Task":

        task = cls(info["task_id"])
        task._info = info
        task._status = models.TaskStatusCode(info["status"])

        # TODO(luispcunha): construct correct output class from API info.

        return task

    @contextlib.contextmanager
    def sync_context(self):
        """Enter context manager for blocking sync execution.

        This turns an asynchronous task into a blocking task.
        If an exception/ctrl+c is caught while in the context manager, the
        remote task is killed.

        Usage:
            task = scenario.simulate(..., run_async=True)

            with task.sync_context():
                # If an exception happens here or ctrl+c is pressed, the task
                # will be killed.
                task.wait()

        """
        try:
            yield
        except KeyboardInterrupt:
            logging.info("Caught SIGINT: terminating blocking task...")
            self.kill()
        except Exception as e:
            logging.error("Caught exception: terminating blocking task...")
            self.kill()
            raise e

    def get_status(self) -> models.TaskStatusCode:
        """Get status of the task.

        This method issues a request to the API.
        """
        # If the task is in a terminal status and we already have the status,
        # return it without refreshing it from the API.
        if self._status is not None and self._status in _TASK_TERMINAL_STATUSES:
            return self._status

        resp = self._api.get_task_status(self._get_path_params())

        return models.TaskStatusCode(resp.body["status"])

    def get_info(self) -> Dict[str, Any]:
        """Get a dictionary with information about the task.

        Information includes e.g., "task_id", "status", timestamps
        ("create_time", "input_submit_time, "start_time", "end_time"),
        among others.

        This method issues a request to the API.
        """
        # If the task is in a terminal status and we already have the info,
        # return it without refreshing it from the API.
        if self._info is not None and self._status in _TASK_TERMINAL_STATUSES:
            return self._info

        params = self._get_path_params()
        resp = self._api.get_task(params, skip_deserialization=True).response

        info = json.loads(resp.data.decode("utf-8"))
        status = models.TaskStatusCode(info["status"])

        self._info = info
        self._status = status

        return info

    def wait(self, polling_period: int = 5) -> models.TaskStatusCode:
        """Wait for the task to complete.

        This method issues requests to the API.

        Args:
            polling_period: How often to poll the API for the task status.

        Returns:
            The final status of the task.
        """
        prev_status = None
        while True:
            status = self.get_status()
            if status != prev_status:
                if status == models.TaskStatusCode.PENDINGINPUT:
                    pass
                elif status == models.TaskStatusCode.SUBMITTED:
                    logging.info("Waiting for resources...")
                elif status == models.TaskStatusCode.STARTED:
                    logging.info("The task is being executed remotely.")
                elif status == models.TaskStatusCode.SUCCESS:
                    logging.info("Task completed successfully.")
                elif status == models.TaskStatusCode.FAILED:
                    logging.info("Task failed.")
                    logging.info("Download the 'stdout.txt' and 'stderr.txt' "
                                 "files with `task.get_output()` for "
                                 "more detail. "
                                 "Post-processing tools will fail.")
                elif status == models.TaskStatusCode.KILLED:
                    logging.info("Task killed.")
                else:
                    logging.info(
                        "An internal error occurred while performing the task.")
            prev_status = status

            if status in _TASK_TERMINAL_STATUSES:
                return status

            time.sleep(polling_period)

    def kill(self) -> None:
        """Kill the task.

        This method issues a request to the API to kill the task. It doesn't
        block waiting for confirmation if the task was killed.
        """
        self._api.kill_task(path_params=self._get_path_params())

    def _get_output_config(self, all_files: bool = False):
        """Get configuration of the output with the task method name.
        
        Args:
            all_files: Whether to download all the files in the output.

        Returns:
            output_class, filenames - If the task method name is not of
                a scenario, this method returns None, None and all
                simulation are downloaded. Otherwise, and if available,
                it returns the output_class of that
                respective scenario and the filenames of that simulation.
                If all_files is False, then the filenames are the default
                files set for that scenario.
        """

        # Fetch the first part of the method_name (e.g., "wind_tunnel")
        method_name = self.get_info()["method_name"].split(".")[0]

        # Set the default files for the output class
        # For some scenarios, there are None default files
        filenames = None
        output_class = None

        if method_name in output_consts.OUTPUT_CONSTS:
            output_class = output_consts.OUTPUT_CONSTS[method_name][
                "output_class"]
            if not all_files:
                filenames = output_consts.OUTPUT_CONSTS[method_name][
                    "default_files"]

        return output_class, filenames

    def get_output(
        self,
        all_files: bool = False,
        output_dir: Optional[pathlib.Path] = None,
        uncompress: bool = True,
        rm_downloaded_zip_archive: bool = True,
    ):
        """Get the output of the task.

        Args:
            all_files: Whether to download all the files in the output
                archive. If False, only the default files are downloaded.
            output_dir: Directory where to download the files. If None, the
                files are downloaded to the default directory. The default is
                {inductiva.working_dir}/{inductiva.output_dir}/{task_id}.
            uncompress: Whether to uncompress the archive after downloading it.
            rm_archive: Whether to remove the archive after uncompressing it.
                If uncompress is False, this argument is ignored.

        Returns:
            Either, return a Scenario output class if the task method name is
            a scenario with an output class, or return an output path otherwise.

        Example:
            task = Task("task_id")
            output_path = task.get_output()
            # prints:
            100%|██████████| 1.64G/1.64G [00:32<00:00, 55.1MiB/s]
        """
        # Get terminal status
        status = self.wait()

        if status == models.TaskStatusCode.SUCCESS:
            output_class, filenames = self._get_output_config(all_files)
        else:
            output_class = None
            logging.info("Downloading the 'stdout.txt' and 'stderr.txt' files.")
            filenames = ["stdout.txt", "stderr.txt"]

        output_dir = self.download_outputs(
            filenames=filenames,
            output_dir=output_dir,
            uncompress=uncompress,
            rm_downloaded_zip_archive=rm_downloaded_zip_archive)

        # output_class can only be not None if the task is successful
        if output_class is not None:
            logging.info("Post-processing tools are available "
                         "through the output object.")
            return output_class(output_dir)

        logging.info("The output was downloaded to %s.", output_dir)
        return output_dir

    def get_output_files_info(self) -> output_contents.OutputContents:
        """Get information of the output files of the task.

        Returns:
            An instance of the OutputContents class, which can be used to
            access info about the output files, such as the archive's size,
            number of files, and information about each file (name, size,
            compressed size). It can also be used to print that information
            in a formatted way.
        """
        api_response = self._api.get_outputs_list(
            path_params=self._get_path_params())

        archive_info = api_response.body

        contents = [
            output_contents.FileInfo(
                name=file_info["name"],
                size=file_info["size"],
                compressed_size=file_info["compressed_size"],
            ) for file_info in archive_info["contents"]
        ]

        return output_contents.OutputContents(
            size=int(archive_info["size"]),
            contents=contents,
        )

    def download_outputs(
        self,
        filenames: Optional[List[str]] = None,
        output_dir: Optional[types.Path] = None,
        uncompress: bool = True,
        rm_downloaded_zip_archive: bool = True,
    ) -> pathlib.Path:
        """Download output files of the task.

        Args:
            filenames: List of filenames to download. If None or empty, all
                files are downloaded.
            output_dir: Directory where to download the files. If None, the
                files are downloaded to the default directory. The default is
                {inductiva.working_dir}/{inductiva.output_dir}/{task_id}.
            uncompress: Whether to uncompress the archive after downloading it.
            rm_downloaded_zip_archive: Whether to remove the archive after
            uncompressing it. If uncompress is False, this argument is ignored.
        """
        api_response = self._api.download_task_output(
            path_params=self._get_path_params(),
            query_params={
                "filename": filenames or [],
            },
            stream=True,
            skip_deserialization=True,
        )
        # use raw urllib3 response instead of the generated client response, to
        # implement our own download logic (with progress bar, first checking
        # the size of the file, etc.)
        response = api_response.response

        if output_dir is None:
            output_dir = files.resolve_path(inductiva.output_dir).joinpath(
                self.id)
        output_dir = files.resolve_path(output_dir)

        if output_dir.exists():
            warnings.warn("Path already exists, files may be overwritten.")
        output_dir.mkdir(parents=True, exist_ok=True)

        zip_path = output_dir.joinpath("output.zip")

        data.download_file(response, zip_path)

        if uncompress:
            data.uncompress_task_outputs(zip_path, output_dir)
            if rm_downloaded_zip_archive:
                zip_path.unlink()

        return output_dir

    class _PathParams(TypedDict):
        """Util class for type checking path params."""
        task_id: str

    def _get_path_params(self) -> _PathParams:
        """Get dictionary with the URL path parameters for API calls."""
        return {"task_id": self.id}

    def get_execution_time(self) -> Optional[float]:
        """Get the time the task took to complete.

        Returns:
            The time in seconds or None if the task hasn't completed yet.
        """
        info = self.get_info()
        if self._status not in _TASK_TERMINAL_STATUSES:
            return None
        # start_time may be None if the task was killed before it started
        if info["start_time"] is None:
            return None

        # Format the time to datetime type
        start_time = datetime.datetime.fromisoformat(info["start_time"])
        end_time = datetime.datetime.fromisoformat(info["end_time"])

        return (end_time - start_time).total_seconds()

    def get_machine_type(self) -> Optional[str]:
        """Get the machine type used in the task.

        Streamlines the process of obtaining the task info, extracting the
        machine type from the comprehensive task info.

        Returns:
            The machine type, or None if a machine hasn't been assigned yet.
        """
        info = self.get_info()
        if info["executer"] is None:
            return None

        machine_info = info["executer"]
        if "vm_type" not in machine_info:
            return "inductiva-machine"

        machine_type = machine_info["vm_type"].split("/")[-1]

        return machine_type



"""Utils to create the material."""

import json


class IsotropicLinearElasticMaterial:
    """Linear elastic isotropic material.

    Attributes:
        young_modulus (float): Young's modulus of the material.
        poisson_ratio (float): Poisson's ratio of the material.
    """

    def __init__(self, young_modulus: float, poisson_ratio: float) -> None:
        """Initializes a IsotropicLinearElasticMaterial object."""
        self.young_modulus = young_modulus
        self.poisson_ratio = poisson_ratio

    def to_dict(self) -> dict:
        """Convert material properties to a dictionary.

        Returns:
            dict: Material properties.
        """
        return {
            "young_modulus": self.young_modulus,
            "poisson_ratio": self.poisson_ratio
        }

    def write_to_json(self, json_path: str) -> None:
        """Writes the material to JSON file.

        Args:
            json_path (str): JSON file path.
        """
        material_dict = self.to_dict()

        # Write JSON file
        with open(json_path, "w", encoding="utf-8") as write_file:
            json.dump(material_dict, write_file, indent=4)


# Define standard materials
STRUCTURAL_STEEL = IsotropicLinearElasticMaterial(young_modulus=200000,
                                                  poisson_ratio=0.3)
ALUMINUM_ALLOY = IsotropicLinearElasticMaterial(young_modulus=71000,
                                                poisson_ratio=0.33)
GRAY_CAST_IRON = IsotropicLinearElasticMaterial(young_modulus=110000,
                                                poisson_ratio=0.28)
MAGNESIUM_ALLOY = IsotropicLinearElasticMaterial(young_modulus=45000,
                                                 poisson_ratio=0.35)
TITANIUM_ALLOY = IsotropicLinearElasticMaterial(young_modulus=96000,
                                                poisson_ratio=0.36)
COPPER_ALLOY = IsotropicLinearElasticMaterial(young_modulus=110000,
                                              poisson_ratio=0.34)



"""GROMACS module of the API"""

from typing import Optional, List

from inductiva import types, tasks, resources, simulators


class GROMACS(simulators.Simulator):
    """Class to invoke any GROMACS command on the API."""

    def __init__(self):
        super().__init__()
        self.api_method_name = "md.gromacs.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        commands: List[dict],
        machine_group: Optional[resources.MachineGroup] = None,
        run_async: bool = False,
    ) -> tasks.Task:
        """Run a list of GROMACS commands.

        Args:
            input_dir: Path to the directory containing the input files.
            commands: List of commands to run using the GROMACS simulator.
            machine_group: The machine group to use for the simulation.
            run_async: Whether to run the simulation asynchronously.
        """
        return super().run(input_dir,
                           machine_group=machine_group,
                           commands=commands,
                           run_async=run_async)



"""Utils to check if optional dependencies are installed.

Example usage, e.g., to check if numpy is installed:

extra_msg = "You can install it with `pip install numpy`."

needs_numpy = functools.partial(
    _needs_optional_deps,
    ["numpy"],
    extra_msg,
)

# Now use `needs_numpy` as a decorator for functions that need numpy.
@optional_deps.needs_numpy
def load_from_np():
    ...
"""

import importlib
import functools


def _needs_optional_deps(module_names, extra_msg, func):
    """Decorator to check if optional dependencies are installed."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        missing_imports = []

        for module_name in module_names:
            try:
                importlib.import_module(module_name)
            except ImportError:
                missing_imports.append(module_name)

        if missing_imports:
            missing_imports_quoted = [f"'{m}'" for m in missing_imports]
            missing_imports_str = ", ".join(missing_imports_quoted)

            msg = (f"Dependencies required to use function "
                   f"'{func.__qualname__}' missing: {missing_imports_str}.")

            if extra_msg:
                msg += f" {extra_msg}"

            raise RuntimeError(msg)

        return func(*args, **kwargs)

    return wrapper


def _missing_deps_msg(extra_deps_name):
    return (f"You can install the missing dependencies for '{extra_deps_name}' "
            f"with `pip install 'inductiva[{extra_deps_name}_extra]'`.")


# Apply the two first arguments to _needs_optional_deps function, creating
# a new decorator function with those arguments already set.
needs_molecules_extra_deps = functools.partial(
    _needs_optional_deps,
    ["MDAnalysis", "nglview", "imageio", "matplotlib"],
    _missing_deps_msg("molecules"),
)

needs_fluids_extra_deps = functools.partial(
    _needs_optional_deps,
    ["pyvista", "vtk", "scipy", "xarray", "dask", "imageio", "matplotlib"],
    _missing_deps_msg("fluids"),
)

needs_coastal_extra_deps = functools.partial(
    _needs_optional_deps,
    ["utm", "imageio", "scipy", "matplotlib"],
    _missing_deps_msg("coastal"),
)

needs_common_extra_deps = functools.partial(
    _needs_optional_deps,
    ["imageio", "matplotlib"],
    _missing_deps_msg("common"),
)



# pylint: disable=unused-argument
"""Deformable plate scenario."""

import os
from typing import List, Optional

from functools import singledispatchmethod

from inductiva import tasks, resources
from inductiva.scenarios import Scenario
from inductiva.simulators import Simulator
from inductiva.simulators import FEniCSx
from inductiva.types import Path

from inductiva.structures import bcs, holes, plates, materials
from . import mesh_utils
from . import bcs_utils
from . import geometry_utils

GEOMETRY_FILENAME = "geometry.json"
MESH_FILENAME = "mesh.msh"
BCS_FILENAME = "bcs.json"
MATERIAL_FILENAME = "material.json"


class DeformablePlate(Scenario):
    """Plate linear elastic scenario.

    The plate linear elastic scenario is characterized by the plate, holes,
    boundary conditions, and material properties. The plate and holes together
    form the geometry, which is used to create the mesh for the simulation.
    """

    valid_simulators = [FEniCSx]

    def __init__(
        self,
        plate: plates.RectangularPlate,
        holes_list: List[holes.Hole],
        bcs_list: List[bcs.BoundaryCondition],
        material: materials.IsotropicLinearElasticMaterial,
    ):
        """Initializes the plate linear elastic scenario.

        Args:
            plate (RectangularPlate): The rectangular plate geometry.
            holes_list (List[Hole]): Holes in the plate.
            bcs_list (List[BoundaryCondition]): The boundary conditions applied
              to the plate.
            material (IsotropicLinearElasticMaterial): The material properties
              of the plate.
            geometry (GeometricCase): The plate with holes geometry.
            bcs_case (BoundaryConditionsCase): The boudnary coinditions for the
             palte with holes.
        """
        self.plate = plate
        self.holes_list = holes_list
        self.bcs_list = bcs_list
        self.material = material
        self.geometry = geometry_utils.GeometricCase(plate=self.plate,
                                                     holes_list=self.holes_list)
        self.bcs_case = bcs_utils.BoundaryConditionsCase(bcs_list=self.bcs_list)

    def simulate(self,
                 simulator: Simulator = FEniCSx(),
                 machine_group: Optional[resources.MachineGroup] = None,
                 run_async: bool = False) -> tasks.Task:
        """Simulates the scenario.

        Args:
            simulator: The simulator to use for the simulation.
            machine_group: The machine group to use for the simulation.
            run_async: Whether to run the simulation asynchronously.
            mesh_filename: Mesh filename.
            bcs_filename: Boundary conditions filename.
            material_filename: Material filename.
        """
        simulator.override_api_method_prefix("deformable_plate")
        task = super().simulate(simulator,
                                machine_group=machine_group,
                                run_async=run_async,
                                mesh_filename=MESH_FILENAME,
                                bcs_filename=BCS_FILENAME,
                                material_filename=MATERIAL_FILENAME)
        return task

    @singledispatchmethod
    def create_input_files(self, simulator: Simulator):
        pass


@DeformablePlate.create_input_files.register
def _(self,
      simulator: FEniCSx,
      input_dir: Path) -> None:
    """Creates FEniCSx simulation input files."""

    # Geometry file
    geometry_path = os.path.join(input_dir, GEOMETRY_FILENAME)
    self.geometry.write_to_json(geometry_path)

    # Mesh file
    mesh = mesh_utils.GmshMesh(self.geometry)
    mesh_path = os.path.join(input_dir, MESH_FILENAME)
    mesh.write_to_msh(mesh_path)

    # BCs file
    bcs_path = os.path.join(input_dir, BCS_FILENAME)
    self.bcs_case.write_to_json(bcs_path)

    # Material file
    material_path = os.path.join(input_dir, MATERIAL_FILENAME)
    self.material.write_to_json(material_path)



"""Describes the physical scenarios and runs its simulation via API."""
from typing import List, Literal, Optional
from enum import Enum
from dataclasses import dataclass

from inductiva import tasks, resources, simulators, fluids


@dataclass
class ParticleRadius(Enum):
    """Sets particle radius according to resolution."""
    HIGH = 0.008
    MEDIUM = 0.012
    LOW = 0.02


class DamBreak(fluids.FluidBlock):
    """Physical scenario of a dam break simulation."""

    def __init__(
        self,
        fluid: fluids.FluidType = fluids.WATER,
        dimensions: Optional[List[float]] = None,
        position: Optional[List[float]] = None,
    ):
        """Initializes a `DamBreak` object.

        Args:
            fluid: A fluid type to simulate.
            dimensions: A list containing fluid column dimensions,
              in meters.
            pisition: A list containing fluid column position in a tank,
              in meters.
            """

        if dimensions is None:
            dimensions = [0.3, 0.3, 0.3]

        super().__init__(density=fluid.density,
                         kinematic_viscosity=fluid.kinematic_viscosity,
                         dimensions=dimensions,
                         position=position)

    # pylint: disable=arguments-renamed
    def simulate(
        self,
        simulator: simulators.Simulator = simulators.DualSPHysics(),
        machine_group: Optional[resources.MachineGroup] = None,
        resolution: Literal["high", "medium", "low"] = "low",
        simulation_time: float = 1,
        run_async: bool = False,
    ) -> tasks.Task:
        """Simulates the scenario.

        Args:
            simulator: Simulator to use.
            machine_group: The machine group to use for the simulation.
            resolution: Resolution of the simulation.
            simulation_time: Simulation time, in seconds.
            run_async: Whether to run the simulation asynchronously.
        """
        simulator.override_api_method_prefix("dam_break")

        self.particle_radius = ParticleRadius[resolution.upper()].value
        self.simulation_time = simulation_time
        self.adaptive_time_step = True
        self.particle_sorting = True
        self.time_step = 0.001
        self.output_time_step = 1 / 60

        # Inherit the simulate from the Parent of FluidBlock (Scenario) to
        # avoid overriding the api_method_prefix with the one of FluidBlock.
        task = super(fluids.FluidBlock, self).simulate(
            simulator=simulator,
            machine_group=machine_group,
            run_async=run_async,
            sim_config_filename=self.get_config_filename(simulator))

        return task



"Postprocessing steps for the MDWaterBox scenario."
import os
import pathlib
from typing import Literal

import numpy as np
try:
    import matplotlib.pyplot as plt
    import nglview as nv
    import IPython as ip
except ImportError:
    nv = None
    plt = None
    ip = None

try:
    import google.colab as gc
except ImportError:
    gc = None

import inductiva
from inductiva.utils import optional_deps

COMPRESSED_TRAJECTORY_FILE = "compressed_trajectory.xtc"
PROTEIN_TOPOLOGY_FILE = "protein.gro"
SYSTEM_TOPOLOGY_FILE = "solvated_protein.tpr"
RMSF_VALUES_FILE = "rmsf_values.npy"


class ProteinSolvationOutput:
    """Post process the simulation output of a ProteinSolvation scenario."""

    def __init__(self, sim_output_path: pathlib.Path = None):
        """Initializes a `ProteinSolvationOutput` object.

        Given a simulation output directory that contains the standard files
        generated by a ProteinSolvation simulation run, this class provides
        methods to visualize the simulation outputs in a notebook interactively.

        Args:
            sim_output_path: Path to the simulation output directory."""

        self.sim_output_dir = sim_output_path
        # universe will be loaded only when needed
        self.universe = None
        self.ipython_kernel = self.get_ipython_kernel()

    @optional_deps.needs_molecules_extra_deps
    def _load_universe(self):
        topology_path = os.path.join(self.sim_output_dir, PROTEIN_TOPOLOGY_FILE)
        trajectory_path = os.path.join(self.sim_output_dir,
                                       COMPRESSED_TRAJECTORY_FILE)

        if self.universe is None:
            self.universe = \
                inductiva.molecules.utils.unwrap_trajectory(
                    topology_path, trajectory_path)

    @optional_deps.needs_molecules_extra_deps
    def render_interactive(self,
                           representation: Literal["cartoon", "ball+stick",
                                                   "line", "point", "surface",
                                                   "ribbon"] = "ball+stick",
                           selection: str = "protein",
                           add_backbone: bool = True):
        """
        Render the simulation outputs in an interactive visualization.
        Args:
            representation: The protein representation to use for the
            visualization.
            selection: The selection to use for the representation. Check
            https://nglviewer.org/ngl/api/manual/usage/selection-language.html
            for details.
            add_backbone: Whether to add the protein backbone to the
            visualization.
            """
        self.enable_vizualization()
        self._load_universe()

        view = nv.show_mdanalysis(self.universe)
        view.add_representation(representation, selection=selection)
        if add_backbone:  #hardcoding the backbone as a cartoon representation
            view.add_representation("cartoon", selection="protein")
        view.center()

        print("System Information: ")
        print(f"Number of protein atoms: {len(self.universe.atoms)}")
        print(f"Number of amino acids: "
              f"{self.universe.select_atoms('protein').n_residues}")
        print(f"Number of trajectory frames: {len(self.universe.trajectory)}")
        return view

    @optional_deps.needs_molecules_extra_deps
    def plot_rmsf_per_residue(self):
        """Plot the root mean square fluctuation (RMSF) over a trajectory.

        It is typically calculated for the alpha carbon atom of each residue.
        These atoms make the backbone of the protein.The RMSF is the square root
        of the variance of the fluctuation around the average position:
        &rhoi = √⟨(xi - ⟨xi⟩)²⟩
        It quantifies how much a structure diverges from the average structure
        over time, the RSMF can reveal which areas of the system are the most
        mobile. Check
        https://userguide.mdanalysis.org/stable/examples/analysis/alignment_and_rms/rmsf.html
        for more details."""
        rmsf_values_path = os.path.join(self.sim_output_dir, RMSF_VALUES_FILE)
        rmsf_values = np.load(rmsf_values_path)
        # Plot the data
        residue_number = np.arange(len(rmsf_values))
        plt.plot(residue_number, rmsf_values)
        plt.xlabel("Residue Number")
        plt.ylabel("RMSF")
        plt.title("RMSF per residue")
        plt.grid(True)
        plt.show()

        return rmsf_values

    @optional_deps.needs_molecules_extra_deps
    def render_attribute_per_residue(
        self,
        residue_attributes: np.ndarray,
        representation: Literal["cartoon", "ball+stick", "line", "point",
                                "surface", "ribbon"] = "cartoon"):
        """Render a specific protein attribute in an interactive visualization.
        Args:
            residue_attributes: The per residue values of the attribute you want
            to visualize.
            representation: The protein representation to use for the
            visualization.
            """
        self.enable_vizualization()
        self._load_universe()

        self.universe.add_TopologyAttr("tempfactors")
        protein = self.universe.select_atoms("protein")
        for residue, value in zip(protein.residues, residue_attributes):
            residue.atoms.tempfactors = value
        view = nv.show_mdanalysis(self.universe)
        view.add_representation(representation, selection="protein")
        view.update_representation(color_scheme="bfactor")
        view.center()
        return view

    @optional_deps.needs_molecules_extra_deps
    def get_ipython_kernel(self):
        """Check if the current environment have an IPython kernel."""
        return ip.get_ipython()

    @optional_deps.needs_molecules_extra_deps
    def enable_vizualization(self):
        """Enable vizualization if IPython is available."""
        if self.ipython_kernel is None:
            raise ImportError("IPython is not available. Visualization is "
                              "only available in a python notebook.")

        if self.ipython_kernel.__module__.startswith("google.colab"):
            gc.output.enable_custom_widget_manager()



"""Post-processing and visualization utilities of the fluid tank scenario."""

import os
from typing import Optional

try:
    import xarray as xr
except ImportError:
    xr = None

from inductiva.types import Path
from inductiva.fluids.post_processing.splishsplash import convert_vtk_data_dir_to_netcdf
from inductiva.utils import files, visualization, optional_deps


class FluidTankOutput:
    """Fluid tank simulation output."""

    def __init__(self, sim_output_path: Path, output_time_step: float = 1):
        """Initializes a `FluidTankOutput` object.

        Args:
            sim_output_path: Path to simulation output files.
        """

        self.sim_output_path = sim_output_path
        self.output_time_step = output_time_step

    @optional_deps.needs_fluids_extra_deps
    def render(
        self,
        movie_path: Path = "movie.mp4",
        fps: int = 10,
        color: Optional[str] = None,
    ):
        """Renders temporal evolution of the fluid particle positions.
        
        A movie is produced representing the temporal evolution of the positions
        of the particles representing the fluid in the tank. Each frame of the
        movie is a scatter plot of the particle positions at a given time.
         
        The movie is saved in the `movie_path` location.

        Args:
            movie_path: Path to the movie file.
            fps: Number of frames per second to use in the movie.
            color: Color used to represent the particles.
        """

        convert_vtk_data_dir_to_netcdf(
            data_dir=os.path.join(self.sim_output_path, "vtk"),
            output_time_step=self.output_time_step,
            netcdf_data_dir=os.path.join(self.sim_output_path, "netcdf"))

        particle_data = xr.open_mfdataset(
            os.path.join(self.sim_output_path, "netcdf", "*.nc"))

        movie_path = files.resolve_path(movie_path)

        visualization.create_3d_scatter_plot_movie(
            particle_data,
            iter_var="time",
            x_var="x",
            y_var="y",
            z_var="z",
            x_limits=[-0.5, 0.5],
            y_limits=[-0.5, 0.5],
            z_limits=[-0.1, 1],
            color=color,
            movie_path=movie_path,
            movie_fps=fps,
        )



"""Class for the coils creation."""
import json
import math
import os
import random
import shutil
import typing

from absl import logging
from functools import singledispatchmethod

import numpy as np

from inductiva import scenarios, simulators, tasks, types, utils, resources

SIMSOPT_COIL_COEFFICIENTS_FILENAME = "coil_coefficients.npz"
SIMSOPT_COIL_CURRENTS_FILENAME = "coil_currents.npz"
SIMSOPT_PLASMA_SURFACE_FILENAME = "input.final"
SIMSOPT_TEMPLATE_DIR = os.path.join(utils.templates.TEMPLATES_PATH,
                                    "stellarator_coils")
PLASMA_SURFACE_TEMPLATE_FILENAME = "input.example"
OBJECTIVES_WEIGHTS_FILENAME = "objectives_weights.json"
DEFAULT_OBJECTIVES_WEIGHTS = {
    "squared_flux": 1,
    "coils_length": 2e-03,
    "mean_squared_curvature": 3e-04,
    "arclength_variation": 5e-03,
    "curvature": 3e-04
}


class StellaratorCoils(scenarios.Scenario):
    """Represents stellarator coils.

    A stellarator is a magnetic fusion confinement device that possesses a 
    set of external electromagnetic coils that have current running through 
    them. These coils create a magnetic field that is able to confine a plasma, 
    at extremely high temperatures, where the nuclear fusion reactions take 
    place.

    Therefore, designing a stellarator involves defining a set of 
    electromagnetic coils. This set of coils is defined by a small set of 
    NC independent coils, due to the symmetries involved. There is stellarator 
    symmetry, that mirrors this independent set of coils and then "number of
    field periods (nfp)" rotational symmetry, that repeats this block of coils
    throughout the stellarator in the toroidal direction (the long way around
    the torus). The number of field periods is the number of complete magnetic 
    field repetitions within a device. It represents how many times the magnetic
    field repeats itself along the toroidal direction, lowering the number
    of independent coils that one has to design. Therefore, the total number
    of coils in a stellarator device comes out to 2*NC*nfp.

    For more information about stellarators, there are many articles with 
    very good information. Per Helander"s 2014 "Theory of plasma confinement in 
    non-axisymmetric magnetic fields" is a good example. See more details at:
    https://iopscience.iop.org/article/10.1088/0034-4885/77/8/087001

    The class can be initialized normally (directly from __init__) providing
    a list of `Coil` objects or from the class methods so that the individual 
    coils can also be created before creating the StellaratorCoils object. 

    Attributes:
        coils (list): List of `Coil` objects.
        num_field_periods (int): Number of magnetic field periods.
          Refers to the number of complete magnetic field repetitions 
          within a stellarator. Represents how many times the magnetic
          field pattern repeats itself along the toroidal direction.
          Typically varies between 1 and 3.
    """

    valid_simulators = [simulators.SIMSOPT]

    def __init__(self, coils, num_field_periods):
        """Initialize the StellaratorCoils object."""

        self.coils = coils
        self.num_field_periods = num_field_periods

    @classmethod
    def from_circular_curves(cls, num_field_periods, num_coils, coil_currents,
                             major_radius, minor_radius):
        """Create simple circular and equally spaced curves.

        The non-zero coefficients are c0 and c1 for both the Fx and Fy
        Fourier Series and s1 for Fz. This is what makes the coils circular. 

        Args:
            num_field_periods (int): Number of magnetic field periods.
            num_coils (int): The number of coils per field period.
            coil_currents (list): List of coil currents.
            major_radius (float): distance from the center of the torus 
              (the central axis) to the outer edge of the plasma region.
            minor_radius (float): Radius of the simple circular curves.

        Returns:
            StellaratorCoils: The created StellaratorCoils instance.
        """

        # To spread the coils equally (in a torus) over an angle that
        # is a fraction of 2*pi. 2*num_field_periods*num_coils will be
        # the final number of coils.
        angle_factor = 2 * num_field_periods * num_coils

        coils = []
        for i in range(num_coils):
            toroidal_angle = (i + 0.5) * (2 * np.pi) / (angle_factor)

            # Set the non-zero coefficients
            curve_coefficients = get_circular_curve_coefficients(
                toroidal_angle, major_radius, minor_radius)

            # Create the Coil object
            coil = Coil(curve_coefficients, coil_currents[i])

            coils.append(coil)

        return cls(coils, num_field_periods)

    @classmethod
    def from_random_curves(cls,
                           num_field_periods,
                           num_coils,
                           coil_currents,
                           max_order=12,
                           major_radius=1.0,
                           minor_radius=0.5):
        """Creates random curves.

        Creates a number of "num_coils" random curves. The random coils are 
        created by first creating simple equally spaced coils and then randomly 
        varying the higher order coefficients of the Fourier series describing 
        the coils. This ensures that the coils are not created on top of each 
        other and are sufficiently separated but with random and complex curves.

        Args:
            num_field_periods (int): Number of magnetic field periods.
            num_coils (int): The number of coils to be randomly created.
            coil_currents (list): List of coil currents.
            max_order (int): Maximum order of the coefficients.
            major_radius (float): distance from the center of the torus 
              (the central axis) to the outer edge of the plasma region.
            minor_radius (float): Radius of the simple initial curves.

        Returns:
            StellaratorCoils: The created StellaratorCoils instance.
        """

        # To spread the coils equally (in a torus) over an angle that
        # is a fraction of 2*pi. 2*num_field_periods*num_coils will be
        # the final number of coils.
        angle_factor = 2 * num_field_periods * num_coils

        coils = []
        for i in range(num_coils):
            toroidal_angle = (i + 0.5) * (2 * np.pi) / (angle_factor)
            curve_coefficients = np.zeros((6, max_order + 1))

            # Base coefficients to make sure the coils are separated.
            curve_coefficients[:, :2] = get_circular_curve_coefficients(
                toroidal_angle, major_radius, minor_radius)

            # Generate random higher coefficients (decreasing the range
            # of the random values to ensure the curve is well behaved).
            for order in range(2, max_order + 1):
                if order == 2:
                    limits = [-0.1, 0.1]

                if 3 <= order <= 5:
                    limits = [-0.05, 0.05]

                if 5 <= order <= 8:
                    limits = [-0.005, 0.005]

                if order > 8:
                    limits = [-0.001, 0.001]

                curve_coefficients[0, order] = random.uniform(*limits)
                curve_coefficients[1, order] = random.uniform(*limits)
                curve_coefficients[2, order] = random.uniform(*limits)
                curve_coefficients[3, order] = random.uniform(*limits)
                curve_coefficients[4, order] = random.uniform(*limits)
                curve_coefficients[5, order] = random.uniform(*limits)

            # Create the Coil object
            coil = Coil(curve_coefficients, coil_currents[i])

            coils.append(coil)

        return cls(coils, num_field_periods)

    @classmethod
    def from_curves_file(cls,
                         num_field_periods,
                         coil_currents,
                         curves_file,
                         delimiter=","):
        """Create StellaratorCoils from Fourier coefficients loaded from a file.

        This function loads a file containing Fourier coefficients for several 
        coils. The file is expected to have `6*num_coils` many columns, and 
        `order+1` many rows. The columns are in the following order,

        sj_x_coil1,cj_x_coil1,sj_y_coil1,...,sj_x_coil2,cj_x_coil2,...

        Args:
            num_field_periods (int): Number of magnetic field periods.
            coil_currents (list): List of coil currents.
            curves_file (str): Name of the file containing Fourier coefficients.
            delimiter (str): Delimiter used in the file. 

        Returns:
            StellaratorCoils: The created StellaratorCoils instance.
        """

        # Reads all the coefficients from the file to a numpy 2D array
        # with `order+1` columns and `6*num_coils` rows.
        curves_data = np.loadtxt(fname=curves_file,
                                 delimiter=delimiter,
                                 unpack=True)

        # Gets the number of coils.
        num_coils = int((curves_data.shape[0]) / 6)

        # Gets the coefficients for each coil.
        curves_coefficients = np.split(curves_data, num_coils, axis=0)

        coils = [
            Coil(curve_coefficients,
                 coil_current) for curve_coefficients, coil_current in zip(
                     curves_coefficients, coil_currents)
        ]

        return cls(coils, num_field_periods)

    def simulate(
        self,
        simulator: simulators.Simulator = simulators.SIMSOPT(),
        machine_group: typing.Optional[resources.MachineGroup] = None,
        run_async: bool = False,
        plasma_surface_filepath: typing.Optional[types.Path] = None,
        num_iterations: int = 1,
        num_samples: int = 1,
        sigma_scaling_factor: float = 0.1,
        objectives_weights: dict = None,
    ) -> tasks.Task:
        """Simulates the scenario.

        The magnetic field produced on the plasma surface and a set of objective
        functions are computed for a collection of stellarator coil 
        configurations with the goal of optimizing a stellarator design.

        The optimization is performed as follows:
        1. The scenario"s coil configuration is used as an initial 
          configuration.
        2. Gaussian noise is added to each coil parameter to produce 
          `num_sample` configurations.
        3. From these configurations, the one with the lowest value of the
          objective functions is selected.
        4. This configuration is then used as an initial configuration for
          the next iteration.
        5. The process is repeated `num_iteration` times.

        The simulation also outputs the Fourier Series coefficients describing
        the coils for each of the configurations obtained during the process.

        Args:
            simulator: The simulator to use for the simulation.
            machine_group: The machine group to use for the simulation.
            run_async: Whether to run the simulation asynchronously.
            plasma_surface_filepath: Path to the file with the description of
              the plasma surface on which the magnetic field will be calculated.
            num_iterations: Number of iterations to run for the searching 
              process.
            num_samples: Number of different stellarator samples generated per 
              iteration from a configuration. The samples are generated using
              normal distribution noise for each coefficient of the Fourier 
              Series that describes the coils.
            sigma_scaling_factor: Scaling factor for the sigma value used in 
              the random generation of noise. This argument makes sure that 
              the noise is generated proportionally to each coefficient. It 
              also determines the range of search for new values of the
              coefficients.
            objectives_weights: Contains the weights for each objective function
              that will be used for the construction of the total objective.
              Only the objectives provided will be used for the calculation.
              Available options for the keys are: "squared_flux", 
              "coils_length", "mean_squared_curvature", "arclength_variation" 
              and "curvature". If `objectives_weights` is not provided it 
              defaults to: {
                                "squared_flux": 1,
                                "coils_length": 2e-03,
                                "mean_squared_curvature": 3e-04,
                                "arclength_variation": 5e-03,
                                "curvature": 3e-04
                            }
                  
        """
        simulator.override_api_method_prefix("stellarators")

        if plasma_surface_filepath:
            self.plasma_surface_filepath = utils.files.resolve_path(
                plasma_surface_filepath)
        else:
            logging.info("Plasma surface description not provided. "
                         "Using default file.")
            self.plasma_surface_filepath = os.path.join(
                SIMSOPT_TEMPLATE_DIR, PLASMA_SURFACE_TEMPLATE_FILENAME)

        # Default weights for the objectives.
        if objectives_weights is None:
            self.objectives_weights = DEFAULT_OBJECTIVES_WEIGHTS

        for key in self.objectives_weights:
            if key not in [
                    "squared_flux", "coils_length", "mean_squared_curvature",
                    "arclength_variation", "curvature"
            ]:
                raise ValueError(
                    "Invalid dictionary keys. "
                    "Available options include: "
                    "`squared_flux`, `coils_length`, "
                    "`mean_squared_curvature`, `arclength_variation` "
                    "and `curvature`.")

        task = super().simulate(
            simulator,
            machine_group=machine_group,
            run_async=run_async,
            coil_coefficients_filename=SIMSOPT_COIL_COEFFICIENTS_FILENAME,
            coil_currents_filename=SIMSOPT_COIL_CURRENTS_FILENAME,
            plasma_surface_filename=SIMSOPT_PLASMA_SURFACE_FILENAME,
            num_field_periods=self.num_field_periods,
            num_iterations=num_iterations,
            num_samples=num_samples,
            sigma_scaling_factor=sigma_scaling_factor,
            objectives_weights_filename=OBJECTIVES_WEIGHTS_FILENAME)

        return task

    @singledispatchmethod
    def create_input_files(self, simulator: simulators.Simulator):
        pass


class Coil:
    """Represents one coil of a stellarator.

    The curve is represented in 3D cartesian coordinates (x, y, z) as a 
    combination of Fourier series [Fx, Fy, Fz], each of which is an expansion 
    over trigonometric functions: 
        Fi = sum_j (sj * sin(j * theta) + cj * cos(j * theta)).
    The curve provided must then be a numpy array with shape `(6, order+1)`, 
    where `order` (the number of columns - 1) is the maximum order of the series 
    (maximum value of j) and the number of rows is 6, one for each of the 
    sj and cj coefficients of each series Fi.

    Attributes: 
        curve_coefficients (np.ndarray): Array with Fourier coefficients 
          defining the coil. Shape: (6, order + 1), where order is the 
          maximum order of the Fourier Series representation.
        current (float): Coil current.
    """

    def __init__(self, curve_coefficients, current):
        """Initialize the Coil object."""

        self.curve_coefficients = curve_coefficients
        self.current = current


def get_circular_curve_coefficients(toroidal_angle, major_radius, minor_radius):
    """Sets the non-zero coefficients of a circular curve.

    Args:
        toroidal_angle (float): Angle that defines the position of the
          coil in the torus.
        major_radius (float): distance from the center of the torus 
          (the central axis) to the outer edge of the plasma region.
        minor_radius (float): Radius of the simple initial curves.

    Returns:
        curve_coefficients (np.ndarray): Array with Fourier coefficients
        defining the curve.
    """
    curve_coefficients = np.zeros((6, 2))

    curve_coefficients[1, 0] = math.cos(toroidal_angle) * major_radius
    curve_coefficients[1, 1] = math.cos(toroidal_angle) * minor_radius
    curve_coefficients[3, 0] = math.sin(toroidal_angle) * major_radius
    curve_coefficients[3, 1] = math.sin(toroidal_angle) * minor_radius
    curve_coefficients[4, 1] = -minor_radius

    return curve_coefficients


@StellaratorCoils.create_input_files.register
def _(self, simulator: simulators.SIMSOPT, input_dir):  # pylint: disable=unused-argument
    """Creates Simsopt simulation input files."""

    coil_coefficients = [coil.curve_coefficients for coil in self.coils]
    coil_currents = np.array([coil.current for coil in self.coils])

    coil_coefficients_filename = os.path.join(
        input_dir, SIMSOPT_COIL_COEFFICIENTS_FILENAME)
    coil_currents_filename = os.path.join(input_dir,
                                          SIMSOPT_COIL_CURRENTS_FILENAME)

    plasma_surface_filename = os.path.join(input_dir,
                                           SIMSOPT_PLASMA_SURFACE_FILENAME)

    np.savez(coil_coefficients_filename, *coil_coefficients)
    np.savez(coil_currents_filename, coil_currents)
    shutil.copy(self.plasma_surface_filepath, plasma_surface_filename)

    # Save the objectives weights dictionary.
    objectives_weights_filepath = os.path.join(input_dir,
                                               OBJECTIVES_WEIGHTS_FILENAME)

    with open(objectives_weights_filepath, "w", encoding="utf-8") as json_file:
        json.dump(self.objectives_weights, json_file)



"""Utilities to generate random 2D arrays with the Diamond-Square algorithm."""

import math
import random
from typing import Sequence

from absl import logging

import numpy as np


def kernel_avg(kernel_offsets,
               random_array,
               center_x,
               center_y,
               kernel_scale,
               enforce_inner_indexes=True):
    """Averaging kernel, given list of relative offsets and scaling factor."""

    # Scale to the distance and then apply offset
    scaled_kernel_offsets = np.multiply(kernel_offsets, kernel_scale)
    src_idxs = scaled_kernel_offsets + [center_x, center_y]

    # Make sure we are only getting indexes inside the map.
    if enforce_inner_indexes:
        all_positive = (src_idxs[:, 0] >= 0) & (src_idxs[:, 1] >= 0)
        max_x, max_y = random_array.shape
        all_bounded = (src_idxs[:, 0] < max_x) & (src_idxs[:, 1] < max_y)
        # Filter out to make sure all indices are inside the map.
        src_idxs = src_idxs[all_positive & all_bounded]

    # Fetch elements by list of x and list of y coords and return the mean.
    elements = random_array[src_idxs[:, 0], src_idxs[:, 1]]
    return np.mean(elements)


def diamond_step(random_array, inter_center_distance, noise_range, random_seed):
    """The diamond step performed at a given inter center distance on a map.
                            +-----------+
                            | D |   | D |
                            |---+---+---|
                            |   | X |   |
                            |---+---+---|
                            | D |   | D |
                            +-----------+
    For each square in the array, set the midpoint to be the average of the four
    corner points plus a random value."""

    diamond_offsets = [[-1, -1], [-1, 1], [1, 1], [1, -1]]
    map_length = random_array.shape[0]
    kernel_scale = inter_center_distance // 2
    random.seed(random_seed)

    for i in range(kernel_scale, map_length, inter_center_distance):
        for j in range(kernel_scale, map_length, inter_center_distance):
            random_array[i, j] = kernel_avg(diamond_offsets, random_array, i, j,
                                            kernel_scale)
            random_array[i, j] += random.uniform(-noise_range, noise_range)


def square_step(random_array, inter_center_distance, noise_range, random_seed):
    """The square step performed at a given inter center distance on a map.
                            +-----------+
                            |   | S |   |
                            |---+---+---|
                            | S | X | S |
                            |---+---+---|
                            |   | S |   |
                            +-----------+
    For each diamond in the array, set the midpoint to be the average of the
    four corner points plus a random value."""

    square_offsets = [[-1, 0], [0, -1], [1, 0], [0, 1]]
    map_length = random_array.shape[0]
    kernel_scale = inter_center_distance // 2
    random.seed(random_seed)
    # Square Step, rows
    for i in range(kernel_scale, map_length, inter_center_distance):
        for j in range(0, map_length, inter_center_distance):
            random_array[i, j] = kernel_avg(square_offsets, random_array, i, j,
                                            kernel_scale)
            random_array[i, j] += random.uniform(-noise_range, noise_range)

    # Square Step, cols
    for i in range(0, map_length, inter_center_distance):
        for j in range(kernel_scale, map_length, inter_center_distance):
            random_array[i, j] = kernel_avg(square_offsets, random_array, i, j,
                                            kernel_scale)
            random_array[i, j] += random.uniform(-noise_range, noise_range)


def iterate_diamond_square(initial_condition: np.ndarray,
                           initial_roughness: float = 1,
                           roughness_factor: float = 0.5,
                           random_seed: int = None):
    """Performs the diamond-square iteration."""

    if initial_condition.shape[0] != initial_condition.shape[1]:
        logging.error(
            'The initial array must be a square. The current shape is: %f x %f',
            initial_condition.shape[0], initial_condition.shape[1])

    logging.info('Generating random array with the Diamond-Square Algorithm...')

    map_length = initial_condition.shape[0]
    inter_center_distance = map_length - 1
    roughness = initial_roughness
    random_array = initial_condition

    while inter_center_distance > 1:

        diamond_step(random_array, inter_center_distance, roughness,
                     random_seed)
        square_step(random_array, inter_center_distance, roughness, random_seed)

        # Go to a finer grid, and decrease the amount of noise
        inter_center_distance //= 2
        roughness *= roughness_factor

    return random_array


def create_initial_condition(size: int,
                             corner_values: Sequence[float]) -> np.ndarray:
    """Creates a square array with the given corner values."""

    initial_condition = np.zeros((size, size), dtype=np.float64)
    corner_idxs = [(0, 0), (0, -1), (-1, 0), (-1, -1)]

    for idx, value in zip(corner_idxs, corner_values):
        initial_condition[idx] = value

    return initial_condition


def create_random_array(size: int,
                        corner_values: Sequence[float],
                        initial_roughness: float = 1,
                        roughness_factor: float = 0.5,
                        random_seed: int = None) -> np.ndarray:
    """Creates a random square array.

    The random array is generated with an implementation of the Diamond-Square
    Algorithm.

    The description of the Diamond-Square algorithms can be found at
    https://en.wikipedia.org/wiki/Diamond-square_algorithm.

    We start with a n x n map, (with n = 2^k + 1) whose 4 corners have certain
    values. Thus, we first apply a diamond step, to fill in the center of that
    map:
        A - - - - - - - - - B                         A - - - - - - - - - B  ---
        - - - - - - - - - - -                         -\\ - - - - - - - //-   ^
        - - - - - - - - - - -                         - -\\ - - - - - //- -   |
        - - - - - - - - - - -                         - - -\\ - - - //- - -   |
        - - - - - - - - - - -                         - - - -\\ - //- - - -   v
        - - - - - - - - - - -   -- diamond step -->   - - - - - X - - - - -  ---
        - - - - - - - - - - -                         - - - -// -\\ - - - -
        - - - - - - - - - - -                         - - -// - - -\\ - - -
        - - - - - - - - - - -                         - -// - - - - -\\ - -
        - - - - - - - - - - -                         -// - - - - - - -\\ -
        C - - - - - - - - - D                         C - - - - - - - - - D
                                                      |<------->|
                                                      kernel_scale
    Then:
        * X = AVG(A, B, C, D)
    AVG is an "averaging kernel". This kernel has a property that we will name
    as kernel_scale. The kernel_scale is the distance (in rows/columns) between
    the point to be computed and the sources (from a diamond or a square).

    Then, we perform a square step. Observe (below) that there are 4 points,
    N, S, W and E we need to fill at the edges of the map (so we can have more
    squares):
                            A - - - > N < - - - B
                            | - - - - ^ - - - - |
                            | - - - - | - - - - |
                            | - - - - | - - - - |
                            v - - - - | - - - - v
                            W < - - - X - - - > E
                            ^ - - - - | - - - - ^
                            | - - - - | - - - - |
                            | - - - - | - - - - |
                            | - - - - v - - - - |
                            C - - - > S < - - - D
    These are computed by considering first the rows (for N and S) and then the
    columns (for W and E). These will be the computations:
        * N = AVG(A, B, X)
        * S = AVG(C, D, X)
        * W = AVG(A, C, X)
        * E = AVG(A, B, X)
    It's important to introduce another concept: inter center distance. This is
    the distance, in number of rows or number of columns, between the points
    that get averaged. For example, in the iteration we have just showed the
    inter center distance the width/height of the map (because we are at the
    first iteration).

    In general:
        inter_center_distance = 2 * kernel_scale + 1

    Conversely:
        kernel_scale = inter_center_distance // 2

    The diamond-square iteration is applied multiple times with increasing
    degrees of detail, which will end up covering the entire map. For that, the
    inter center distance is halved at every new diamond-square iteration until
    we get to a distance of 1. At that point, all elements of the map grid
    have been generated.

    This averaging procedure has a natural smoothing effect that ensures
    continuity in the terrain. There is, however, some randomness being added at
    every moment, to ensure we don't get a simple smooth profile. However, the
    amount of noise introduced at every step of the iteration allow us to get
    both realistic randomness and, at the same time, realistic continuity.

    Args:
        size: Array size. Must be 2^k + 1. The shape of the array is (size,
          size).
        initial_roughness: Roughness applied in the first iteration of the
          Diamond-Square algorithm. The roughness is a measure of the amount of
          randomness that is added to the array in each step of the algorithm.
        roughness_factor: Fractional value in range (0, 1) by which the initial
          roughness is multiplied at every iteration of the Diamond-Square
          algorithm. Determines how smooth the final array is. Smaller values
          lead to smoother arrays.
        random_seed: Random seed used to generate the random array.
    Returns:
        Random array.
    """

    if not math.log2(size - 1).is_integer():
        raise ValueError('`size` must be 2^k + 1.')

    if len(corner_values) != 4:
        raise ValueError('`corner_values` must be a sequence of length 4, '
                         'containing the values of the corners of the array.')

    initial_condition = create_initial_condition(size, corner_values)

    random_array = iterate_diamond_square(initial_condition, initial_roughness,
                                          roughness_factor, random_seed)

    return random_array



"""Functions to manage or retrieve user resources."""
from absl import logging
import inductiva
import inductiva.client
from inductiva.client.apis.tags import instance_api
from inductiva.utils import format_utils


def _machine_group_list_to_str(machine_group_list) -> str:
    """Returns a string representation of a list of machine groups."""
    columns = [
        "Name", "VM Type", "# machines", "Disk Size in GB", "Spot", "Started at"
    ]
    rows = []

    for machine_group in machine_group_list:
        rows.append([
            machine_group.name, machine_group.machine_type,
            machine_group.num_machines, machine_group.disk_size_gb,
            machine_group.spot, machine_group.create_time
        ])

    formatters = {"Started at": format_utils.datetime_formatter}
    override_col_space = {
        "VM Type": 15,
        "# machines": 12,
        "Spot": 10,
    }

    return format_utils.get_tabular_str(
        rows,
        columns,
        default_col_space=18,
        override_col_space=override_col_space,
        formatters=formatters,
    )


def _fetch_machine_groups_from_api():
    """Get all active machine groups of a user from the API."""
    try:
        api = instance_api.InstanceApi(inductiva.api.get_client())
        response = api.list_active_user_instance_groups()
        if len(response.body) == 0:
            logging.info("No active machine groups found.")
            return response.body

        return response.body

    except inductiva.client.ApiException as api_exception:
        raise api_exception


# pylint: disable=redefined-builtin
def list():
    # pylint: disable=line-too-long
    """Lists all active machine groups info.

    This method queries Google Cloud for active machine groups
    that belong to the user. It outputs all the information relative to
    each machine group as folllows:

    INFO:absl:Active machine groups:
                                        Name         VM Type   # machines    Disk Size in GB       Spot         Started at
    api-1b1f724c-5cfe-4d87-8439-9689aa139723   c2-standard-4            1                 40      False   13 Sep, 07:38:50
    api-8e6bf7d8-4888-4de9-bda5-268484b46e6f   c2-standard-4            1                 40      False   13 Sep, 07:37:49

    The name of the machine group can be used to retrieve a MachineGroup object
    with the 'get' function."""
    # pylint: enable=line-too-long

    machine_group_list = get()
    if len(machine_group_list) != 0:
        print("Active machine groups:")
        print(_machine_group_list_to_str(machine_group_list))


def get():
    """Returns a list of 'MachineGroup' objects."""

    # Retrive the active machine group names
    machine_groups = _fetch_machine_groups_from_api()

    machine_group_list = [
        inductiva.resources.MachineGroup.from_api_response(mg) \
            for mg in machine_groups
    ]

    return machine_group_list



"""Tools to analyze molecular dynamics simulations."""
from typing import Optional

try:
    import MDAnalysis as mda
    from MDAnalysis import transformations
    from MDAnalysis.analysis import align
except ImportError:
    mda = None
    transformations = None
    align = None

from inductiva.utils import optional_deps, files


@optional_deps.needs_molecules_extra_deps
def unwrap_trajectory(topology, trajectory):
    """Unwrap visualization of the trajectory to deal with
    Periodic Boundary Conditions.
    Args:
        topology: Path to the topology file.
        trajectory: Path to the trajectory file."""
    universe = mda.Universe(topology, trajectory, guess_bonds=True)
    atoms = universe.atoms
    transformation = transformations.unwrap(atoms)
    universe.trajectory.add_transformations(transformation)
    return universe


@optional_deps.needs_molecules_extra_deps
def align_trajectory_to_average(universe, trajectory_output_path):
    """Align the trajectory to the average structure.
    Args:
        universe: The universe MDAnalysis object.
        trajectory_output_path: Path to the aligned trajectory file."""
    average = align.AverageStructure(universe,
                                     universe,
                                     select="protein and name CA",
                                     ref_frame=0).run()
    average_trajectory = average.results.universe

    align.AlignTraj(universe,
                    average_trajectory,
                    select="protein and name CA",
                    filename=trajectory_output_path,
                    in_memory=False).run()


def download_pdb_from_rcsb(pdb_id: str, save_dir: Optional[str] = None) -> str:
    """Download a PDB file from the RCSB database according
    to its pdb ID.
    Args:
        pdb_id: The PDB identifier of the molecule to download.
        save_dir: The directory to save the file to. If None
            is passed, this will download to the current working
            directory. If the save_dir passed does not exist, this
            method will try to create it.
    Returns:
        The path to the downloaded PDB file.
    """
    pdb_url = f"https://files.rcsb.org/download/{pdb_id}.pdb"

    file_path = files.download_from_url(pdb_url, save_dir)

    return file_path



"""Post-processing and visualization utilities of the coastal area scenario."""

import os
import tempfile
from typing import Dict, Literal, Optional, Sequence, Tuple
from absl import logging

import numpy as np
try:
    import matplotlib.pyplot as plt
    import scipy
except ImportError:
    plt = None
    scipy = None

from inductiva.types import Path
from inductiva.utils import visualization, optional_deps

# Labels used in SWASH simulation outputs.
QUANTITY_SWASH_LABELS = {
    "water_level": "WATLEV",
    "velocity_x": "VKSI",
    "velocity_y": "VETA",
    "velocity_magnitude": "VMAG",
}

# Labels used in plots.
QUANTITY_PLOT_LABELS = {
    "water_level": "Water level",
    "velocity_x": "Velocity x-component",
    "velocity_y": "Velocity y-component",
    "velocity_magnitude": "Velocity magnitude",
}

# Units used in plots.
QUANTITY_UNITS = {
    "water_level": "m",
    "velocity_x": "m/s",
    "velocity_y": "m/s",
    "velocity_magnitude": "m/s",
}

GRID_POSITIONS_FILE_NAME = "grid_positions.mat"


@optional_deps.needs_coastal_extra_deps
def read_swash_output_file(file_path: str) -> Dict[str, np.ndarray]:
    """Reads a SWASH output file.

    The output is read as a dictionary, where the keys are the names of the
    variables and the values are the corresponding arrays.
    """

    try:
        # Read file contents
        data_dict = scipy.io.loadmat(file_path)

    except:
        raise ValueError(f"Error reading file '{file_path}'.") from None

    return data_dict


@optional_deps.needs_coastal_extra_deps
def read_swash_grid_positions_file(file_path: str) -> Tuple[np.ndarray]:
    """Reads a SWASH output file containing grid positions."""

    # Read grid positions
    grid_positions_dict = read_swash_output_file(file_path)

    # Get grids with x and y coordinates
    x_grid = grid_positions_dict["Xp"].transpose()
    y_grid = grid_positions_dict["Yp"].transpose()

    # Get arrays with x and y coordinates
    x_array = x_grid[:, 0]
    y_array = y_grid[0, :]

    return x_array, y_array


@optional_deps.needs_coastal_extra_deps
def read_swash_grid_quantity_file(file_path: str, quantity: str):
    """Reads a SWASH output file containing a quantity defined over a grid.

    SWASH outputs are stored in .mat files. Here, we first read the contents
    of one of these files as a dictionary, and then convert the dictionary
    to a list of times and a list of data arrays.

    # The list of times is built from the keys of the dictionary:
    # the keys end with a number, in format 123456_789, identifying the time
    # instant 12 hours, 34 minutes, 56 seconds and 789 milliseconds; we
    # strip the times from the keys and convert them to seconds.

    # The list of data arrays is built from the values of the dictionary:
    # the values correspond to the data values at each of the time instants
    # represented by the corresponding keys; we read each value as a 2D
    # tensor with indices (x, y).
    """

    # Read data file
    data_dict = read_swash_output_file(file_path)

    # Convert data dictionary to lists of times and data arrays
    time_list = []
    data_list = []

    for key, data_values in data_dict.items():
        # Check if key starts with the quantity label attributed by SWASH
        if key.startswith(QUANTITY_SWASH_LABELS[quantity].capitalize()):
            # Parse key string to get time in milliseconds
            time_ms = key[len(QUANTITY_SWASH_LABELS[quantity]):].replace(
                "_", "")

            # Convert time to seconds
            time_s = float(time_ms[:2]) * 3600. + \
                     float(time_ms[2:4]) * 60. + \
                     float(time_ms[4:6]) + \
                     float(time_ms[6:]) / 1000.

            # The values are transposed in the file, so we transpose them
            # again for consistency
            transposed_data_values = data_values.transpose()

            time_list.append(time_s)
            data_list.append(transposed_data_values)

    return time_list, data_list


@optional_deps.needs_coastal_extra_deps
def _render_quantity_grid_data(x_array: np.ndarray,
                               y_array: np.ndarray,
                               time_list: Sequence[float],
                               data_list: Sequence[np.ndarray],
                               quantity: str,
                               movie_path: str,
                               fps: int = 10,
                               cmap: Optional[str] = None,
                               clim: Optional[Sequence[float]] = None):
    """Renders temporal evolution of a quantity defined over a grid."""

    with tempfile.TemporaryDirectory() as tmp_dir:

        axes_limits = [
            x_array.min(),
            x_array.max(),
            y_array.min(),
            y_array.max(),
        ]

        if clim is None:
            clim = [
                np.min([np.min(data) for data in data_list]),
                np.max([np.max(data) for data in data_list]),
            ]

        for idx, (time, data) in enumerate(zip(time_list, data_list)):
            fig = plt.figure()
            ax = fig.add_subplot()

            im = ax.imshow(
                np.transpose(data),
                origin="lower",
                cmap=cmap,
                clim=clim,
                extent=axes_limits,
                aspect="equal",
            )

            plt.colorbar(
                im,
                ax=ax,
                label=(f"{QUANTITY_PLOT_LABELS[quantity]} "
                       f"[{QUANTITY_UNITS[quantity]}]"),
            )

            ax.set(
                xlabel="x [m]",
                ylabel="y [m]",
                title=f"Time: {time:.2f} [s]",
            )

            fig.savefig(os.path.join(tmp_dir, f"{idx:06d}.png"))
            plt.close(fig)

        visualization.create_movie_from_frames(frames_dir=tmp_dir,
                                               movie_path=movie_path,
                                               fps=fps)


class CoastalAreaOutput:
    """Heat sink simulation output."""

    def __init__(self, sim_output_path: Path):
        """Initializes a `HeatSinkOutput` object.

        Args:
            sim_output_path: Path to simulation output files.
        """

        self.sim_output_path = sim_output_path
        self.check_stability()

    @optional_deps.needs_coastal_extra_deps
    def check_stability(self):

        _, data_list = read_swash_grid_quantity_file(
            file_path=os.path.join(self.sim_output_path, "water_level.mat"),
            quantity="water_level",
        )

        # Check if water level is stable
        inital_max_water_level = np.max(data_list[0])
        final_max_water_level = np.max(data_list[-1])

        if final_max_water_level > 2 * inital_max_water_level:
            logging.info("Simulation can show unstable results.\n"
                         "Check the simulation visualization to corroborate.\n"
                         "If the results are unstable, try increasing "
                         "the simulation resolution.")

    @optional_deps.needs_coastal_extra_deps
    def render(
        self,
        quantity: Literal[
            "water_level",
            "velocity_x",
            "velocity_y",
            "velocity_magnitude",
        ] = "water_level",
        movie_path: Path = "movie.mp4",
        fps: int = 5,
        cmap: str = "viridis",
        clim: Optional[Sequence[float]] = None,
    ):
        """Renders temporal evolution of a physical quantity.

        A movie is produced representing the temporal evolution of a physical
        quantity exported in simulations of the coastal area scenario. Available
        quantities are: water_level, velocity_x, velocity_y, velocity_magnitude.

        The movie is saved in the `movie_path` location.

        Args:
            quantity: Quantity to render. Available quantities are: water_level,
              velocity_x, velocity_y, velocity_magnitude.
            movie_path: Path to the movie file.
            fps: Number of frames per second to use in the movie.
            cmap: Colormap used to represent the quantity.
            clim: Colorbar limits. When not specified, the limits are
              automatically determined from the data.
        """

        x_array, y_array = read_swash_grid_positions_file(
            os.path.join(self.sim_output_path, GRID_POSITIONS_FILE_NAME))

        time_list, data_list = read_swash_grid_quantity_file(
            file_path=os.path.join(self.sim_output_path, quantity + ".mat"),
            quantity=quantity,
        )

        logging.info("Starting to render video from simulation data...")

        _render_quantity_grid_data(
            x_array=x_array,
            y_array=y_array,
            time_list=time_list,
            data_list=data_list,
            quantity=quantity,
            movie_path=movie_path,
            fps=fps,
            cmap=cmap,
            clim=clim,
        )

        logging.info("Writing mp4 file to %s.", movie_path)



"""Util functions for formatting data for printing to console."""
import datetime
import pandas as pd
import numpy as np


def bytes_formatter(n_bytes: int) -> str:
    """Convert bytes to human readable string."""
    res = float(n_bytes)

    for unit in ["B", "KiB", "MiB", "GiB", "TiB"]:
        if res < 1024:
            if unit == "B":
                return f"{res:.0f} {unit}"
            else:
                return f"{res:.2f} {unit}"
        res /= 1024

    return f"{bytes_formatter:.2f} PiB"


def datetime_formatter(dt: str) -> str:
    return datetime.datetime.fromisoformat(dt).strftime("%d %b, %H:%M:%S")


def seconds_formatter(secs: float) -> str:
    return (f"{int(secs // 3600)}h "
            f"{int((secs % 3600) // 60)}m "
            f"{int(secs % 60)}s")


def get_tabular_str(
    rows,
    columns,
    default_col_space=15,
    override_col_space=None,
    formatters=None,
) -> str:
    """Converts a list of lists to a string table.

    Temporary solution to display tables. `pandas` dependency to print tables
    is overkill, but since we have `pandas` anyway for now, we can use it.
    """
    df = pd.DataFrame(rows, columns=columns)

    formatters = formatters or {}

    col_space = {col: default_col_space for col in columns}
    col_space.update(override_col_space or {})

    # replace None with np.nan so that pandas can format them as "n/a"
    # by passing na_rep="n/a" to to_string()
    df.fillna(np.nan, inplace=True)

    return df.to_string(
        index=False,
        na_rep="n/a",
        formatters=formatters,
        col_space=col_space,
    )



"""MachineGroup class to manage Google Cloud resources."""
import time

from absl import logging

import inductiva
import inductiva.client.models
from inductiva import api
from inductiva.client.apis.tags import instance_api


class MachineGroup():
    """Class to manage Google Cloud resources."""

    def __init__(
        self,
        machine_type: str,
        num_machines: int = 1,
        spot: bool = False,
        disk_size_gb: int = 40,
        zone: str = "europe-west1-b",
        register: bool = True,
    ) -> None:
        """Create a MachineGroup object.

        Args:
            machine_type: The type of GC machine to launch. Ex: "e2-medium".
            Check https://cloud.google.com/compute/docs/machine-resource for
            more information about machine types.
            num_machines: The number of virtual machines to launch.
            spot: Whether to use spot machines.
            disk_size_gb: The size of the disk in GB, recommended min. is 40 GB.
            zone: The zone where the machines will be launched.
        """
        self.id = None
        self.name = None
        self.create_time = None
        self.machine_type = machine_type
        self.num_machines = num_machines
        self.spot = spot
        self.disk_size_gb = disk_size_gb
        self.zone = zone
        self._started = False

        # Set the API configuration that carries the information from the client
        # to the backend.
        self._api = instance_api.InstanceApi(api.get_client())
        self._estimated_cost = None

        if register:
            self._register_machine_group()

    def _register_machine_group(self):
        instance_group_config = inductiva.client.models.InstanceGroupCreate(
            machine_type=self.machine_type,
            num_instances=self.num_machines,
            spot=self.spot,
            disk_size_gb=self.disk_size_gb,
            zone=self.zone,
        )
        logging.info("Registering machine group configuration with the API.")
        resp = self._api.register_instance_group(body=instance_group_config)
        self.id = resp.body["id"]
        self.name = resp.body["name"]
        logging.info("Registered machine group:")
        logging.info(" > ID: %s", self.id)
        logging.info(" > Name: %s", self.name)

    @classmethod
    def from_api_response(cls, resp: dict):
        """Creates a MachineGroup object from an API response."""

        machine_group = cls(
            machine_type=resp["machine_type"],
            num_machines=int(resp["num_instances"]),
            spot=bool(resp["spot"]),
            disk_size_gb=resp["disk_size_gb"],
            zone=resp["zone"],
            register=False,
        )
        machine_group.id = resp["id"]
        machine_group.name = resp["name"]
        machine_group.create_time = resp["create_time"]
        machine_group._started = True

        return machine_group

    def start(self):
        """Starts a machine group."""
        if self._started:
            logging.info("Attempting to start a machine group already started.")
            return

        if self.id is None or self.name is None:
            logging.info("Attempting to start an unregistered machine group. "
                         "Make sure you have called the constructor without "
                         "`register=False`.")
            return

        request_body = \
            inductiva.client.models.InstanceGroup(
                id=self.id,
                name=self.name,
                machine_type=self.machine_type,
                num_instances=self.num_machines,
                spot=self.spot,
                disk_size_gb=self.disk_size_gb,
                zone=self.zone,
            )
        try:
            logging.info("Starting a machine group. "
                         "This may take up to a few minutes.")
            logging.info("Note that stopping this local process will not "
                         "interrupt the creation of the machine group. "
                         "Please wait...")
            self._started = True
            start_time = time.time()
            self._api.start_instance_group(body=request_body)
            creation_time_mins = (time.time() - start_time) / 60

            logging.info("Machine group successfully started in %.2f mins.",
                         creation_time_mins)
            self._log_machine_group_info()

        except inductiva.client.ApiException as api_exception:
            raise api_exception

    def terminate(self):
        """Terminates a machine group."""
        if not self._started or self.id is None or self.name is None:
            logging.info("Attempting to terminate an unstarted machine group.")
            return

        try:
            logging.info("Terminating machine group. "
                         "This may take up to a few minutes.")
            start_time = time.time()

            request_body = \
                inductiva.client.models.InstanceGroup(
                    id=str(self.id),
                    name=self.name,
                    machine_type=self.machine_type,
                    num_instances=self.num_machines,
                    spot=self.spot,
                    disk_size_gb=self.disk_size_gb,
                    zone=self.zone,
                )

            self._api.delete_instance_group(body=request_body)
            termination_time_mins = (time.time() - start_time) / 60
            logging.info(
                "Machine group of %s machines successfully "
                "terminated in %.2f mins.", self.num_machines,
                termination_time_mins)

        except inductiva.client.ApiException as api_exception:
            raise api_exception

    def _get_estimated_cost(self) -> float:
        if self._estimated_cost is not None:
            return self._estimated_cost
        instance_price = self._api.get_instance_price({
            "machine_type": self.machine_type,
            "zone": self.zone,
        })
        if self.spot:
            estimated_cost = instance_price.body[
                "preemptible_price"] * self.num_machines
        else:
            estimated_cost = instance_price.body[
                "on_demand_price"] * self.num_machines

        self._estimated_cost = float(round(estimated_cost, 3))
        return self._estimated_cost

    def estimate_cloud_cost(self) -> float:
        """Returns the estimated cost per hour of a machine group.

        Note that this is an estimate of the cost incurred in cloud resources,
        and is not binding of the actual price of the machine group.
        """
        #TODO: Contemplate disk size in the price.
        estimated_cost = self._get_estimated_cost()
        logging.info("Estimated cloud cost per hour: %s $/h", estimated_cost)
        return estimated_cost

    def status(self):
        """Returns the status of a machine group if it exists.

        Otherwise returns None"""
        if self.name is None:
            logging.info(
                "Attempting to get the status of an unregistered machine "
                "group.")
            return

        response = self._api.get_group_status({"name": self.name})

        if response.body == "notFound":
            logging.info("Machine group does not exist: %s.", self.name)
        return response.body

    def _log_machine_group_info(self):
        """Logs the machine group info."""

        logging.info("Name: %s", self.name)
        logging.info("Machine type: %s", self.machine_type)
        logging.info("Number of machines: %s", self.num_machines)
        logging.info("Spot: %s", self.spot)
        logging.info("Disk size: %s GB", self.disk_size_gb)
        logging.info("Estimated cost per hour: %s $/h",
                     self._get_estimated_cost())



"""Interpolation utilities."""

from typing import Literal, Optional, Tuple

import numpy as np
try:
    import scipy
except ImportError:
    scipy = None


def interpolate_to_coordinates(
    coordinates: Tuple[np.ndarray],
    values: np.ndarray,
    interpolation_coordinates: Tuple[np.ndarray],
    method: Literal["nearest", "linear"] = "linear",
    threshold_distance: Optional[float] = None,
):
    """Interpolate values defined at coordinates x to interpolation coordinates.

    Args:
        coordinates: Tuple of arrays with the coordinates of the points where
          values are defined.
        values: Array with the values defined at the points defined by `x`.
        interpolation_coordinates: Tuple of arrays with the coordinates of the
          points where values are interpolated to.
        threshold_distance: If not `None`, points in `interpolation_coordinates`
          that are farther than `threshold_distance` from the closest point in
          `coordinates` are set to `np.NAN`.
    
    Returns:
        Array with the interpolated values. Its shape matches that of the
        interpolation coordinate arrays.
    """

    if method == "nearest":
        interpolator = scipy.interpolate.NearestNDInterpolator(
            coordinates, values)
    elif method == "linear":
        interpolator = scipy.interpolate.LinearNDInterpolator(
            coordinates, values)
    else:
        raise ValueError("Invalid interpolation method.")

    values_interpolation = interpolator(interpolation_coordinates)

    if threshold_distance is not None:
        # Filter out points that are far from x.
        tree = scipy.spatial.KDTree(np.transpose(coordinates))

        # Obtain distance between each interpolation point and the closest
        # location where values are defined.
        interpolation_coordinates_shape = np.shape(interpolation_coordinates)

        # Queries to the KDTree require that interpolation coordinates are
        # provided in shape num_points x dim_coordinates, where num_points
        # is the number of points to query and dim_coordinates is the number
        # of coordinates per point.
        # Reshape interpolation coordinates to match that shape.
        interpolation_coordinates_reshaped = np.reshape(
            interpolation_coordinates, (interpolation_coordinates_shape[0], -1))
        interpolation_coordinates_reshaped = np.transpose(
            interpolation_coordinates_reshaped)
        distance, _ = tree.query(interpolation_coordinates_reshaped, k=1)
        distance = distance.reshape(interpolation_coordinates_shape[1:])

        # Set values to NaN for points that are far from locations where values
        # are defined.
        values_interpolation[distance > threshold_distance] = np.nan

    return values_interpolation



"""Represent a complex terrain profile."""
import typing

try:
    import pyvista as pv
except ImportError:
    pv = None

import inductiva
from inductiva.generative import procedural
from inductiva.utils import optional_deps


class Terrain:
    """Represents a Terrain profile.
    
    A Terrain profile is defined by the surface of the terrain.
    Here, we describe it by a mesh with coordinates (x, y, z).
    The (x, y) reflect the (longitude, latitude) of the terrain
    and the z coordinate is the elevation of the terrain.
    """

    def __init__(self, terrain_mesh):
        """Initializes the `Terrain` object.
        
        Attributes:
            terrain_mesh (pyvista.StructuredGrid): Terrain mesh."""
        self.mesh = terrain_mesh

    @classmethod
    def from_file(cls, terrain_file: str):
        """Creates a `Terrain` object from a text file."""
        terrain_mesh = pv.read(terrain_file)

        return cls(terrain_mesh)

    @classmethod
    @optional_deps.needs_fluids_extra_deps
    def from_random_generation(cls,
                               x_range: typing.Sequence[float],
                               y_range: typing.Sequence[float],
                               x_num: int,
                               y_num: int,
                               corner_values: typing.List[float],
                               initial_roughness: float = 1,
                               roughness_factor: float = 0.5,
                               random_seed: int = None):
        """Creates a `Terrain` object with random elevations.
        The elevation of the corners are randomly selected within
        the range of 0 to 1. This limits the maximum terrain height
        due to the nature of the terrain generation algorithm.
        To allow for more complex terrains, we allow the user to multiply
        by a `heigh_factor` to increase the maximum height of the terrain.
        In this way, the maximum_height possible will be around 
        the value of the `height_factor`.

        TODO: Improve terrain generation for high discrepancy
        terrains.

        Args:
            x_range = [x_min, x_max]: Range of the x-axis in meters (m).
            y_range = [y_min, y_max]: Range of the y-axis in meters (m).
            x_num: Number of points in the x-axis.
            y_num: Number of points in the y-axis.
            corner_values: List of the corner values of the terrain.
            initial_roughness: Initial roughness of the terrain.
            roughness_factor: Factor to multiply the roughness by.
            min_elevation: Minimum elevation of the terrain.
            max_elevation: Maximum elevation of the terrain.
            random_seed: Random seed to use for the terrain generation.
        """

        map_level = procedural.generate_random_map_level(
            x_num=x_num,
            y_num=y_num,
            corner_values=corner_values,
            initial_roughness=initial_roughness,
            roughness_factor=roughness_factor,
            random_seed=random_seed)

        x_grid, y_grid = inductiva.utils.grids.get_meshgrid(x_range=x_range,
                                                            y_range=y_range,
                                                            x_num=x_num,
                                                            y_num=y_num)

        terrain = pv.StructuredGrid(x_grid, y_grid, map_level)

        return cls(terrain)

    @optional_deps.needs_fluids_extra_deps
    def to_text_file(self, text_file_path: str):
        """Saves the terrain to a text file.
        
        Args
            text_file_path: Path to the text file
        """

        terrain_geometry = self.mesh.extract_geometry()
        terrain_geometry.save(text_file_path)

    @property
    @optional_deps.needs_fluids_extra_deps
    def bounds(self):
        """Returns the bounds of the terrain."""

        terrain_bounds = {
            "x": [self.mesh.bounds[0], self.mesh.bounds[1]],
            "y": [self.mesh.bounds[2], self.mesh.bounds[3]],
            "z": [self.mesh.bounds[4], self.mesh.bounds[5]]
        }

        return terrain_bounds

    @property
    @optional_deps.needs_fluids_extra_deps
    def center(self):
        """Returns the center of the terrain."""

        terrain_center = {
            "x": (self.mesh.bounds[0] + self.mesh.bounds[1]) / 2,
            "y": (self.mesh.bounds[2] + self.mesh.bounds[3]) / 2
        }

        return terrain_center

    @property
    @optional_deps.needs_fluids_extra_deps
    def mesh_resolution(self):
        """Returns the resolution of the terrain mesh."""

        num_cells = self.mesh.n_cells
        num_points = self.mesh.n_points

        return num_cells, num_points

    @optional_deps.needs_fluids_extra_deps
    def plot(self,
             off_screen: bool = False,
             save_path: str = None,
             show_edges: bool = False,
             background_color: str = "white",
             colormap: str = "cividis",
             lighting: bool = True,
             camera: typing.List[tuple] = None):
        """Renders the terrain.
        
        Args:
            off_screen: If True, the plot is not shown.
            save_path: Path to save the screenshot.
            show_edges: If True, the edges of the mesh are shown.
            background_color: Background color of the plot.
            colormap: Colormap to use.
            lighting: Enable or disable view direction lighting.
                It highlights the details in the terrain.
            camera: List of tuples given as 
                [position, focal_point, view_up] that describe
                the position of the camera, the focal point it focus
                on and the (negative) direction of the camera.
                Defaults to None, which finds the best isometric
                view based on the provided mesh.
        """

        # Create Elevation from z-values to plot by it.
        self.mesh["Elevation"] = self.mesh.z.T.reshape(-1)
        plotter = pv.Plotter(off_screen=off_screen)
        plotter.camera_position = camera

        # Plot the grid on the back of the terrain.
        plotter.show_grid(mesh=self.mesh, location="outer")
        plotter.background_color = background_color
        plotter.add_mesh(self.mesh,
                         show_edges=show_edges,
                         scalars="Elevation",
                         lighting=lighting,
                         cmap=colormap)

        plotter.show(screenshot=save_path)
        plotter.close()



"""Client for Inductiva's web API."""
import os
import logging

import absl

from . import api
from . import output_consts
from . import simulators
from . import fluids
from . import coastal
from . import scenarios
from . import generative
from . import molecules
from . import resources
# from . import structures
from . import templates
from . import utils
from . import plasmas
from . import world
from . import tasks

api_url = os.environ.get("INDUCTIVA_API_URL", "http://api.inductiva.ai")
output_dir = os.environ.get("INDUCTIVA_OUTPUT_DIR", "inductiva_output")
api_key = os.environ.get("INDUCTIVA_API_KEY")
working_dir = None

# Disable urllib3 warnings.
# TODO: Verify and fix the appearance of this warning.
urllib3_logger = logging.getLogger("urllib3.connectionpool")
urllib3_logger.setLevel(logging.CRITICAL)

absl.logging.set_verbosity(absl.logging.INFO)



"""Coastal area scenario."""

from functools import singledispatchmethod
import math
import os
import shutil
from typing import Literal, Optional
import numpy as np

from absl import logging

from inductiva import tasks, resources, simulators, scenarios, utils
from inductiva import coastal

SCENARIO_TEMPLATE_DIR = os.path.join(utils.templates.TEMPLATES_PATH,
                                     "coastal_area")
SWASH_TEMPLATE_SUBDIR = "swash"
SWASH_CONFIG_TEMPLATE_FILENAME = "input.sws.jinja"
SWASH_CONFIG_FILENAME = "input.sws"
SWASH_BATHYMETRY_FILENAME = "bathymetry.bot"


class CoastalArea(scenarios.Scenario):
    """Coastal area scenario.

    This is a simulation scenario for waves propagating in a coastal area
    represented by an arbitrary bathymetric profile (i.e., the map of depths of
    the sea bottom).

    The scenario is simulated in a 2D box (x points east, y points north) with
    dimensions defined by the bathymetric profile. Waves are injected from one
    of the boundaries of the simulation domain with a given amplitude and
    period. The base water level is also configurable.

    The remaining three boundaries of the simulation domain are absorbing.
    Absorption in these boundaries may not be perfect, so small reflections may
    be observed.

    Schematic representation of the simulation scenario: x points right, y
    points up.
    _________________________________
    |                               |
    |                               |
    |                               |
    |  injected waves ->     beach  |
    |                               |
    |                               |
    |_______________________________|
    Note that waves may be injected from other boundaries, and that the beach
    may be located elsewhere in the simulation domain.

    The scenario can be simulated with SWASH.
    """

    valid_simulators = [simulators.SWASH]

    def __init__(
        self,
        bathymetry: coastal.Bathymetry,
        water_level: float = 0,
        wave_source_location: Literal["N", "S", "E", "W"] = "W",
        wave_amplitude: float = 2,
        wave_period: float = 10,
    ):
        """Initializes a `CoastalArea` object.

        Args:
            bathymetry: The bathymetric profile.
            water_level: The water level, in meters.
            wave_source_location: The location of the wave source. Supported
              locations are: N (north), S (south), E (east), W (west),
              corresponding to the upper, lower, right and left boundaries of
              the simulation domain, respectively.
            wave_amplitude: The amplitude of the wave, in meters.
            wave_period: The period of the wave, in seconds.
        """

        if not bathymetry.is_uniform_grid():
            logging.info("The bathymetry is not defined on a uniform grid. "
                         "Attempting to interpolate it to a uniform grid...")
            bathymetry = bathymetry.to_uniform_grid()

        self.bathymetry = bathymetry
        self.water_level = water_level
        self.wave_source_location = wave_source_location
        self.wave_amplitude = wave_amplitude
        self._check_valid_wave()
        self.wave_period = wave_period

    def simulate(
        self,
        simulator: simulators.Simulator = simulators.SWASH(),
        machine_group: Optional[resources.MachineGroup] = None,
        run_async: bool = False,
        simulation_time: float = 100,
        time_step: float = 0.1,
        output_time_step: float = 1,
    ) -> tasks.Task:
        """Simulates the scenario.

        Args:
            simulator: Simulator to use. Supported simulators are: SWASH.
            machine_group: The machine group to use for the simulation.
            simulation_time: Total simulation time, in seconds.
            time_step: Time step, in seconds.
            output_time_step: Time step for the output, in seconds.
        """
        simulator.override_api_method_prefix("coastal_area")

        self.simulation_time = simulation_time
        self.time_step = time_step
        self.output_time_step = output_time_step

        task = super().simulate(
            simulator,
            machine_group=machine_group,
            run_async=run_async,
            sim_config_filename=SWASH_CONFIG_FILENAME,
        )

        return task

    @singledispatchmethod
    def get_config_filename(self, simulator: simulators.Simulator):
        pass

    @singledispatchmethod
    def create_input_files(self, simulator: simulators.Simulator):
        pass

    def _check_valid_wave(self):
        """Checks that the wave amplitude is valid.

        Raises:
            ValueError: If the wave amplitude is not valid.
        """

        depths_grid = self.bathymetry.depths_grid()

        if self.wave_source_location == "W":
            wave_boundary = depths_grid[0, :]

        elif self.wave_source_location == "E":
            wave_boundary = depths_grid[-1, :]

        elif self.wave_source_location == "N":
            wave_boundary = depths_grid[:, -1]

        elif self.wave_source_location == "S":
            wave_boundary = depths_grid[:, 0]

        else:
            raise ValueError("Invalid wave source location. "
                             "Supported locations are: N, S, E, W")

        #The greater the value of depth, the deeper the water, so
        #we want to check the minimum depth using the min value

        minimum_depth = np.min(wave_boundary)
        if self.wave_amplitude > minimum_depth:
            raise ValueError("Wave amplitude is larger than the "
                             "minimum depth of the bathymetry in the "
                             "boundary where waves are generated. \n"
                             "This can be due to a very high wave "
                             "amplitude or the waves are being generated "
                             "on land")


@CoastalArea.get_config_filename.register
def _(cls, simulator: simulators.SWASH):  # pylint: disable=unused-argument
    """Returns the configuration filename for SWASH."""
    return SWASH_CONFIG_FILENAME


@CoastalArea.create_input_files.register
def _(self, simulator: simulators.SWASH, input_dir):  # pylint: disable=unused-argument
    """Creates SPlisHSPlasH simulation input files."""

    template_files_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                      SWASH_TEMPLATE_SUBDIR)

    shutil.copytree(template_files_dir,
                    input_dir,
                    dirs_exist_ok=True,
                    symlinks=True)

    config_template_file_path, config_file_path = (os.path.join(
        input_dir, file_name
    ) for file_name in [SWASH_CONFIG_TEMPLATE_FILENAME, SWASH_CONFIG_FILENAME])

    # Compute the bathymetry grid spacing.
    bathymetry_x_num = len(self.bathymetry.x_uniques())
    bathymetry_y_num = len(self.bathymetry.y_uniques())

    bathymetry_x_delta = (self.bathymetry.x_ptp()) / bathymetry_x_num
    bathymetry_y_delta = (self.bathymetry.y_ptp()) / bathymetry_y_num

    # SWASH requires the simulation time to be formatted as HHMMSS.sss.
    simulation_time_hmsms = _convert_time_to_hmsms(self.simulation_time)

    # SWASH uses as amplitude the peak-to-peak amplitude.
    wave_amplitude = 2 * self.wave_amplitude

    # All boundaries except the wave source are absorbing.
    absorbing_boundary_locations = ["N", "S", "E", "W"]
    absorbing_boundary_locations.remove(self.wave_source_location)

    utils.templates.replace_params(
        template_path=config_template_file_path,
        params={
            "bathymetry_filename": SWASH_BATHYMETRY_FILENAME,
            "x_range": self.bathymetry.x_range,
            "y_range": self.bathymetry.y_range,
            "x_num": bathymetry_x_num - 1,
            "y_num": bathymetry_y_num - 1,
            "x_delta": bathymetry_x_delta,
            "y_delta": bathymetry_y_delta,
            "water_level": self.water_level,
            "wave_source_location": self.wave_source_location,
            "wave_amplitude": wave_amplitude,
            "wave_period": self.wave_period,
            "absorbing_boundary_locations": absorbing_boundary_locations,
            "simulation_time_hmsms": simulation_time_hmsms,
            "time_step": self.time_step,
            "output_time_step": self.output_time_step,
        },
        output_file=config_file_path,
        remove_template=True,
    )

    bathymetry_file_path = os.path.join(input_dir, SWASH_BATHYMETRY_FILENAME)
    depths_grid = self.bathymetry.depths.reshape(
        (bathymetry_x_num, bathymetry_y_num))

    self.bathymetry.to_bot_file(bathymetry_file_path, depths_grid)


def _convert_time_to_hmsms(time: float) -> str:
    """Converts time in seconds to hours, minutes, seconds and milliseconds.

    The output format is HHMMSS.sss, where HH, MM, SS and sss are the
    zero-padded values of hours, minutes, seconds and milliseconds in the time,
    respectively.
    """

    time_m, time_s = divmod(time, 60.)
    time_h, time_m = divmod(time_m, 60.)
    time_ms, time_s = math.modf(time_s)

    return (f"{int(time_h):02d}"
            f"{int(time_m):02d}"
            f"{int(time_s):02d}."
            f"{int(time_ms*1000):03d}")



"""Utils to create the different types of boundary conditions."""

from typing import List

import json

from inductiva.structures import bcs


class BoundaryConditionsCase:
    """Boundary conditions for a given case.

    The boundary conditions for a given case are characterized by a list of 
      Dirichlet and Neumann boundary conditions.

    Attributes:
        bcs_list (list): The boundary conditions objects.
    """

    def __init__(self, bcs_list: List[bcs.BoundaryCondition]) -> None:
        """Initializes a BoundaryConditionsCase object."""
        self.bcs_list = bcs_list

    def write_to_json(self, json_path: str) -> None:
        """Write the boundary conditions to JSON file.

        Args:
            json_path (str): The JSON file path.
        """

        # Divide the bcs_obj_list into two lists based on their types
        dirichlet_bcs = [
            bc for bc in self.bcs_list if isinstance(bc, bcs.DirichletBC)
        ]
        neumann_bcs = [
            bc for bc in self.bcs_list if isinstance(bc, bcs.NeumannBC)
        ]

        # Dirichlet dictionary
        if dirichlet_bcs is not None:
            list_of_dirichlet_bc_dicts = []
            for dirichlet_bc in dirichlet_bcs:
                dirichlet_bc_dict = dirichlet_bc.to_dict()
                list_of_dirichlet_bc_dicts.append(dirichlet_bc_dict)
            dirichlet_dict = {"dirichlet": list_of_dirichlet_bc_dicts}

            # Merge dictionaries: boundary conditions dictionary
            bcs_dict = {**dirichlet_dict}

        # Neumann dictionary
        if neumann_bcs is not None:
            list_of_neumann_bcs_dicts = []
            for neumann_bc in neumann_bcs:
                neumann_bc_dict = neumann_bc.to_dict()
                list_of_neumann_bcs_dicts.append(neumann_bc_dict)
            neumann_dict = {"neumann": list_of_neumann_bcs_dicts}

            # Merge dictionaries: boundary conditions dictionary
            bcs_dict = {**bcs_dict, **neumann_dict}

        # Write JSON file
        with open(json_path, "w", encoding="utf-8") as write_file:
            json.dump(bcs_dict, write_file, indent=4)



#pylint: disable=missing-module-docstring
from .fluid_types import (
    FluidType,
    WATER,
    OLIVE_OIL,
    LIQUID_PROPANE,
    JET_FUEL,
    GEAR_OIL,
    BEER,
    HONEY,
)

from .post_processing import SteadyStateOutput
from .wind_tunnel import WindTunnel, WindTunnelOutput
from .fluid_block import FluidBlock
from .dam_break import DamBreak
from .heat_sink import HeatSink, HeatSinkOutput
from .wind_terrain import WindOverTerrain
from . import shapes
from .fluid_tank import (FluidTank, CubicTankOutlet, CylindricalTankOutlet,
                         CircularTankInlet, FluidTankOutput)
from ._post_processing import SPHSimulationOutput

from . import post_processing



"""SplisHSPlasH simulator module of the API."""
from typing import Optional

from inductiva import types, tasks, resources
from inductiva.simulators import Simulator


class SplishSplash(Simulator):
    """Class to invoke a generic SPlisHSPlasH simulation on the API."""

    def __init__(self):
        super().__init__()
        self.api_method_name = "sph.splishsplash.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        sim_config_filename: str,
        machine_group: Optional[resources.MachineGroup] = None,
        run_async: bool = False,
        particle_radius: float = 0.025,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            sim_config_filename: Name of the simulation configuration file.
            other arguments: See the documentation of the base class.
        """
        return super().run(
            input_dir,
            machine_group=machine_group,
            input_filename=sim_config_filename,
            run_async=run_async,
            particle_radius=particle_radius,
        )



"""DualSPHysics simulator module of the API."""

from typing import Optional

from inductiva import types, tasks, resources
from inductiva.simulators import Simulator


class DualSPHysics(Simulator):
    """Class to invoke a generic DualSPHysics simulation on the API."""

    def __init__(self):
        super().__init__()
        self.api_method_name = "sph.dualsphysics.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        sim_config_filename: str = "config.xml",
        machine_group: Optional[resources.MachineGroup] = None,
        run_async: bool = False,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            sim_config_filename: Name of the simulation configuration file.
            other arguments: See the documentation of the base class.
        """
        return super().run(input_dir,
                           machine_group=machine_group,
                           input_filename=sim_config_filename,
                           run_async=run_async)



"""Fluid tank scenario."""
from dataclasses import dataclass
import enum
from functools import singledispatchmethod
import json
import os
from typing import List, Literal, Optional

from inductiva import resources, fluids, simulators, scenarios, tasks, utils

SCENARIO_TEMPLATE_DIR = os.path.join(utils.templates.TEMPLATES_PATH,
                                     "fluid_tank")
SPLISHSPLASH_TEMPLATE_INPUT_DIR = "splishsplash"
SPLISHSPLASH_TEMPLATE_FILENAME = "fluid_tank_template.splishsplash.json.jinja"
SPLISHSPLASH_CONFIG_FILENAME = "fluid_tank.json"
TANK_JSON_FILENAME = "tank.json"
TANK_MESH_FILENAME = "tank.obj"
FLUID_MESH_FILENAME = "fluid.obj"


@dataclass
class ParticleRadius(enum.Enum):
    """Sets particle radius according to resolution."""
    HIGH = 0.005
    MEDIUM = 0.01
    LOW = 0.02


@dataclass
class TimeStep(enum.Enum):
    """Sets time step according to resolution."""
    HIGH = 0.0025
    MEDIUM = 0.005
    LOW = 0.01


# Tank inlets.
class BaseTankInlet:
    """Base tank inlet."""

    def __init__(self, fluid_velocity: float):
        """Initializes a base tank inlet.

        Args:
            fluid_velocity: Fluid velocity.
        """
        self.fluid_velocity = fluid_velocity

    def to_dict(self) -> dict:
        """Returns a dictionary representation of the inlet."""
        return {
            "fluid_velocity": self.fluid_velocity,
        }


class CircularTankInlet(BaseTankInlet):
    """Circular tank inlet."""

    def __init__(self,
                 fluid_velocity: float = 1,
                 position: List[float] = (0, 0),
                 radius: float = 0.1):
        """Initializes a circular tank inlet."""
        self.shape = fluids.shapes.Circle(radius=radius, position=position)
        super().__init__(fluid_velocity=fluid_velocity)

    def to_dict(self) -> dict:
        """Returns a dictionary representation of the inlet."""
        return {
            **super().to_dict(),
            "shape": self.shape.to_dict(),
        }


# Tank outlets.
class BaseTankOutlet:
    """Base tank outlet."""
    pass


class CubicTankOutlet(BaseTankOutlet):
    """Cubic tank outlet."""

    def __init__(
        self,
        dimensions: List[float],
        top_base_position: Optional[List[float]] = None,
    ):
        """Initializes a cubic tank outlet.

        Args:
            dimensions: Dimensions of the outlet.
            top_base_position: Position of the top base of the outlet.
        """
        top_base_position = top_base_position or [0, 0]

        self.shape = fluids.shapes.Cube(
            dimensions=dimensions,
            position=[*top_base_position, -dimensions[2]],
        )

    def to_dict(self) -> dict:
        """Returns a dictionary representation of the outlet."""
        return {"shape": self.shape.to_dict()}


class CylindricalTankOutlet(BaseTankOutlet):
    """Cylindrical tank outlet."""

    def __init__(
        self,
        radius: float,
        height: float,
        top_base_position: Optional[List[float]] = None,
    ):
        """Initializes a cylindrical tank outlet.

        Args:
            radius: Radius of the outlet.
            height: Height of the outlet.
            top_base_position: Position of the top base of the outlet.
        """
        top_base_position = top_base_position or [0, 0]

        self.shape = fluids.shapes.Cylinder(
            radius=radius,
            height=height,
            position=[*top_base_position, -height],
        )

    def to_dict(self) -> dict:
        """Returns a dictionary representation of the outlet."""
        return {"shape": self.shape.to_dict()}


class FluidTank(scenarios.Scenario):
    """Fluid tank scenario.

    This is a simulation scenario for a fluid tank. The tank has a 3D shape that
    may be cubic or cylindrical. Fluid is injected in the tank via an inlet
    located at the top of the tank, and flows out of the tank via an outlet
    located at the bottom of the tank. The motion of the fluid is controled
    by gravity. The fluid properties such as density and kinematic viscosity
    are configurable. The initial fluid level in the tank is also configurable,
    as well as the inlet and outlet positions and dimensions.

    The main axis of the tank is the z axis. The inlet and outlet are located at
    the top and bottom of the tank, respectively, along the z axis. Fluid is
    injected from the inlet in the negative z direction with a given velocity.

    Schematic representation of the simulation scenario: e.g. x/y points right,
    z points up.

       inlet
    _____________________
    |    |              |
    |    v              |
    |                   |
    |                   |
    |___________________|  fluid level
    |                   |
    |                   |
    |                   |
    |                   |
    |___________________|
                    |
                    v  outlet

    The scenario can be simulated with SPlisHSPlasH.
    """

    valid_simulators = [simulators.SplishSplash]

    def __init__(
        self,
        shape: fluids.shapes.BaseShape = fluids.shapes.Cylinder(radius=0.5,
                                                                height=1),
        fluid: fluids.FluidType = fluids.WATER,
        fluid_level: float = 0.5,
        inlet: Optional[BaseTankInlet] = CircularTankInlet(radius=0.1),
        outlet: Optional[BaseTankOutlet] = CylindricalTankOutlet(radius=0.1,
                                                                 height=0.1),
    ):
        """Initializes a fluid tank.

        Args:
            shape: The shape of the tank.
            fluid: The fluid type.
            fluid_level: The fluid level initially in the tank.
            inlet: The inlet of the tank.
            outlet: The outlet of the tank.
        """
        self.shape = shape
        self.fluid = fluid
        self.fluid_level = fluid_level
        self.inlet = inlet
        self.outlet = outlet

    def simulate(
        self,
        simulator: simulators.Simulator = simulators.SplishSplash(),
        machine_group: Optional[resources.MachineGroup] = None,
        run_async: bool = False,
        simulation_time: float = 5,
        resolution: Literal["low", "medium", "high"] = "low",
        output_time_step: float = 0.1,
        particle_sorting: bool = False,
    ) -> tasks.Task:
        """Simulates the scenario.

        Args:
            simulator: Simulator to use. Supported simulators are: SPlisHSPlasH.
            machine_group: The machine group to use for the simulation.
            simulation_time: Total simulation time, in seconds.
            output_time_step: Time step for the output, in seconds.
            resolution: Resolution of the simulation. Controls the particle
              radius and time step. Accepted values are: "low", "medium",
              "high".
            particle_sorting: Whether to use particle sorting.
            run_async: Whether to run the simulation asynchronously.
        """
        simulator.override_api_method_prefix("fluid_tank")

        self.simulation_time = simulation_time
        self.particle_radius = ParticleRadius[resolution.upper()].value
        self.time_step = TimeStep[resolution.upper()].value
        self.output_time_step = output_time_step
        self.particle_sorting = particle_sorting

        task = super().simulate(
            simulator,
            machine_group=machine_group,
            run_async=run_async,
            particle_radius=self.particle_radius,
            sim_config_filename=self.get_config_filename(simulator),
        )

        return task

    def get_bounding_box(self):
        """Gets the bounding box of the tank.

        Returns:
            Tuple of two lists representing the minimum and maximum coordinates
            of the bounding box of the tank, respectively.
        """

        bounding_box_min, bounding_box_max = self.shape.get_bounding_box()

        # Extend the bounding box to include the outlet.
        if self.outlet is not None:
            outlet_bounding_box_min, _ = self.outlet.shape.get_bounding_box()
            bounding_box_min[2] = outlet_bounding_box_min[2]

        return bounding_box_min, bounding_box_max

    def to_dict(self) -> dict:
        """Returns a dictionary representation of the scenario."""
        return {
            "shape": self.shape.to_dict(),
            "fluid": self.fluid.to_dict(),
            "fluid_level": self.fluid_level,
            "inlet": self.inlet.to_dict(),
            "outlet": self.outlet.to_dict(),
        }

    def create_json_file(self, output_path):
        """Creates a JSON file with the scenario parameters."""

        with open(output_path, "w", encoding="utf-8") as file:
            json.dump(self.to_dict(), file)

    @singledispatchmethod
    def get_config_filename(self, simulator: simulators.Simulator):
        pass

    @singledispatchmethod
    def create_input_files(self, simulator: simulators.Simulator):
        pass


@FluidTank.get_config_filename.register
def _(cls, simulator: simulators.SplishSplash) -> str:  # pylint: disable=unused-argument
    """Returns the config filename for SPlisHSPlasH."""
    return SPLISHSPLASH_CONFIG_FILENAME


@FluidTank.create_input_files.register
def _(self, simulator: simulators.SplishSplash, input_dir):  # pylint: disable=unused-argument
    """Creates SPlisHSPlasH simulation input files."""

    template_files_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                      SPLISHSPLASH_TEMPLATE_INPUT_DIR)
    self.create_json_file(os.path.join(input_dir, TANK_JSON_FILENAME))

    bounding_box_min, bounding_box_max = self.get_bounding_box()
    inlet_position = [
        self.inlet.shape.position[0],
        self.inlet.shape.position[1],
        bounding_box_max[2],
    ]

    utils.templates.replace_params(
        template_path=os.path.join(template_files_dir,
                                   SPLISHSPLASH_TEMPLATE_FILENAME),
        params={
            "simulation_time": self.simulation_time,
            "time_step": self.time_step,
            "particle_radius": self.particle_radius,
            "data_export_rate": 1 / self.output_time_step,
            "z_sort": self.particle_sorting,
            "tank_filename": TANK_MESH_FILENAME,
            "fluid_filename": FLUID_MESH_FILENAME,
            "fluid": self.fluid,
            "inlet_position": inlet_position,
            "inlet_width": int(self.inlet.shape.radius / self.particle_radius),
            "inlet_fluid_velocity": self.inlet.fluid_velocity,
            "bounding_box_min": bounding_box_min,
            "bounding_box_max": bounding_box_max,
        },
        output_file=os.path.join(input_dir, SPLISHSPLASH_CONFIG_FILENAME),
    )
